from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from transformers import PreTrainedTokenizer

from owlsight.utils.logger import logger


class TextGenerationProcessor(ABC):
    """Abstract base class for text generation processors implementing basic generation."""

    def __init__(
        self,
        model_id: str,
        save_history: bool,
        system_prompt: str,
    ):
        """
        Initialize the text generation processor.

        Parameters
        ----------
        model_id : str
            The model ID to use for generation.
            Usually the name of the model or the path to the model.
        save_history : bool
            Whether or not to save the history of inputs and outputs.
        system_prompt : str
            The system prompt to use for generation.
        """
        if not model_id:
            raise ValueError("Model ID cannot be empty.")

        self.model_id = model_id
        self.save_history = save_history
        self.system_prompt = system_prompt
        self.history = []

    def apply_chat_template(
        self,
        input_data: str,
        tokenizer: PreTrainedTokenizer,
    ) -> str:
        """
        Apply chat template to the input text.
        This is used to format the input text before generating a response and should be universal across all models.

        Parameters
        ----------
        input_data : str
            The input text to apply the template to.
        tokenizer : PreTrainedTokenizer
            The tokenizer to use for applying the template.

        Returns
        -------
        str
            The formatted text with the chat template applied.
        """
        if tokenizer.chat_template is not None:
            messages = self.get_history()
            messages.append({"role": "user", "content": input_data})
            templated_text = tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        else:
            logger.warning("Chat template not found in tokenizer. Using input text as is.")
            templated_text = input_data

        return templated_text

    def update_history(self, input_data: str, generated_text: str) -> None:
        """
        Update the history with the input and generated text.

        Parameters
        ----------
        input_data : str
            The input text that was provided.
        generated_text : str
            The text that was generated in response.
        """
        if self.save_history:
            self.history.append({"role": "user", "content": input_data})
            self.history.append({"role": "assistant", "content": generated_text.strip()})

    def get_history(self) -> List[Dict[str, str]]:
        """
        Get complete chat history of inputs and outputs and system prompt.

        Returns
        -------
        List[Dict[str, str]]
            The chat history including system prompt if present.
        """
        messages = self.history.copy()
        if self.system_prompt:
            messages.insert(0, {"role": "system", "content": self.system_prompt})
        return messages

    @abstractmethod
    def generate(
        self,
        input_data: str,
        max_new_tokens: int,
        temperature: float,
        generation_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> str:
        """Generate text based on input data."""
        raise NotImplementedError("generate method must be implemented in the subclass.")



import pytest
import sys
import re
import json
from copy import deepcopy
from unittest.mock import patch, mock_open

sys.path.append("src")
from owlsight.configurations.config_manager import ConfigManager, DottedDict, _prepare_toggle_choices
from owlsight.utils.constants import DEFAULTS, ConfigSchema


@pytest.fixture(scope="function", autouse=True)
def defaults():
    """Fixture to reset DEFAULTS after each test."""
    yield ConfigSchema.get_defaults()

@pytest.fixture(scope="function", autouse=True)
def choices():
    """Fixture to return the choices."""
    yield ConfigSchema.get_choices()

@pytest.fixture
def config_manager():
    """Fixture to return a new instance of ConfigManager."""
    return ConfigManager()


def test_singleton_behavior():
    """Ensure ConfigManager follows the singleton pattern."""
    instance1 = ConfigManager()
    instance2 = ConfigManager()
    assert instance1 is instance2, "ConfigManager is not a singleton!"


def test_get_existing_key(config_manager):
    """Test the retrieval of an existing config key."""
    value = config_manager.get("main.max_retries_on_error")
    assert value == 3, f"Expected 3, got {value}"


def test_get_non_existing_key(config_manager):
    """Test getting a non-existent key."""
    value = config_manager.get("non.existing.key", default="default_value")
    assert value == "default_value", f"Expected 'default_value', got {value}"


def test_set_new_key(config_manager):
    """Test setting a new config key."""
    config_manager.set("new.key", "new_value")
    assert config_manager.get("new.key") == "new_value", "Failed to set new key!"


def test_set_existing_key(config_manager):
    """Test setting an existing config key."""
    config_manager.set("main.max_retries_on_error", 5)
    assert config_manager.get("main.max_retries_on_error") == 5, "Failed to update existing key!"


@patch("builtins.open", new_callable=mock_open)
def test_save_config(mock_file, config_manager):
    """Test saving configuration to a file."""
    with patch("os.path.exists", return_value=True):
        result = config_manager.save("test_config.json")
        assert result, "Save operation should return True on success"
        mock_file.assert_called_once_with("test_config.json", "w")
        mock_file().write.assert_called()


@patch("builtins.open", new_callable=mock_open, read_data=json.dumps(DEFAULTS))
def test_load_config(mock_file, config_manager):
    """Test loading configuration from a file."""
    with patch("os.path.exists", return_value=True):
        result = config_manager.load("test_config.json")
        assert result, "Load operation should return True on success"
        mock_file.assert_called_once_with("test_config.json", "r")
        assert config_manager.get("main.max_retries_on_error") == 3, "Failed to load correct config value"


@patch("os.path.exists", return_value=False)
@patch("owlsight.configurations.config_manager.logger")
def test_load_non_existing_file(mock_logger, mock_exists, config_manager):
    """Test loading a non-existent config file."""
    result = config_manager.load("non_existent_config.json")
    assert not result, "Load operation should return False on failure"
    mock_logger.error.assert_called_with(
        "Cannot load config. Configuration file does not exist: 'non_existent_config.json'"
    )


@patch("os.path.exists", return_value=True)
@patch("builtins.open", new_callable=mock_open, read_data='{"invalid": "json"')
def test_load_invalid_json(mock_file, mock_exists, config_manager):
    """Test loading an invalid JSON file."""
    result = config_manager.load("invalid_config.json")
    assert not result, "Load operation should return False on invalid JSON"


@patch("os.path.exists", return_value=True)
@patch("builtins.open", new_callable=mock_open, read_data='{"invalid_key": "value"}')
def test_load_invalid_config(mock_file, mock_exists, config_manager):
    """Test loading a config with invalid keys."""
    result = config_manager.load("invalid_config.json")
    assert not result, "Load operation should return False on invalid config"


def test_validate_config_missing_sections(config_manager):
    """Test config validation with missing keys."""
    invalid_config = {"main": {}}
    expected_missing_sections = set(DEFAULTS.keys()) - set(invalid_config.keys())
    expected_error_message = f"Config misses the following sections: {expected_missing_sections}"

    with pytest.raises(KeyError, match=re.escape(expected_error_message)):
        config_manager.validate_config(invalid_config)


def test_validate_config_invalid_keys(config_manager, defaults):
    """Test config validation with invalid sections."""
    invalid_config = {**defaults, "invalid_section": {}}
    expected_invalid_sections = {"invalid_section"}
    expected_error_message = (
        f"Config has the following sections, which are not valid in owlsight: {expected_invalid_sections}"
    )

    with pytest.raises(KeyError, match=re.escape(expected_error_message)):
        config_manager.validate_config(invalid_config)


def test_validate_config_invalid_value_type(config_manager, defaults):
    """Test config validation with invalid value types."""
    invalid_config = {**defaults, "main": {**defaults["main"], "extra_index_url": 5}}
    with pytest.raises(TypeError, match="Invalid type int for key 'main.extra_index_url'. Expected type: str"):
        config_manager.validate_config(invalid_config)


def test_validate_config_invalid_choice(config_manager, defaults):
    """Test config validation with invalid choice value."""
    invalid_config = {**defaults, "main": {**defaults["main"], "max_retries_on_error": 100}}
    with pytest.raises(ValueError, match="Invalid value 100 for key 'main.max_retries_on_error'. Possible values:"):
        config_manager.validate_config(invalid_config)


def test_dotted_dict():
    """Test DottedDict functionality."""
    dotted = DottedDict({"key": "value", "nested": {"inner_key": "inner_value"}})
    assert dotted.key == "value", "DottedDict failed to retrieve a top-level key"
    assert dotted.nested.inner_key == "inner_value", "DottedDict failed to retrieve a nested key"

    # Test case insensitivity
    assert dotted.KEY == "value", "DottedDict should be case-insensitive"

    # Test setting and deleting attributes
    dotted.new_key = "new_value"
    assert dotted.new_key == "new_value", "DottedDict failed to set a new key"
    del dotted.new_key
    assert "new_key" not in dotted, "DottedDict failed to delete a key"


def test_config_choices(config_manager, defaults, choices):
    """Test the config choices are generated correctly."""
    choices = config_manager.config_choices
    assert "main" in choices, "Main config missing from config choices"
    assert "max_retries_on_error" in choices["main"], "Max retries on error choice missing"
    assert choices["main"]["max_retries_on_error"] == _prepare_toggle_choices(
        defaults["main"]["max_retries_on_error"], choices["main"]["max_retries_on_error"]
    ), "Invalid toggle choices for max retries on error"


def test_prepare_toggle_choices():
    """Test the _prepare_toggle_choices function."""
    current_val = 3
    possible_vals = [1, 2, 3, 4, 5]
    result = _prepare_toggle_choices(current_val, possible_vals)
    assert result == [3, 4, 5, 1, 2], "Toggle choices not prepared correctly"

    # Test when current_val is not in possible_vals
    current_val = 6
    result = _prepare_toggle_choices(current_val, possible_vals)
    assert result == possible_vals, "Toggle choices should remain unchanged when current_val is not in possible_vals"


def test_copy_defaults_does_not_modify_original(defaults):
    """Test that the DEFAULTS copy does not modify the original DEFAULTS."""
    assert defaults["main"]["max_retries_on_error"] == 3, "DEFAULTS should have the right default value"
    defaults_copy = deepcopy(defaults)
    defaults_copy["main"]["max_retries_on_error"] = 5
    assert defaults["main"]["max_retries_on_error"] != 5, "DEFAULTS should not be modified"


def test_defaults_variable_should_not_be_modified(config_manager, defaults):
    """Test that the DEFAULTS variable is not modified by the ConfigManager."""
    config_manager.set("main.max_retries_on_error", 5)
    assert defaults["main"]["max_retries_on_error"] != 5, "ConfigManager should not modify DEFAULTS"


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from typing import Optional, Dict, Any, Union, List
import traceback
from pathlib import Path
import io
import requests
import numpy as np
from PIL import Image
import re

from owlsight.hugging_face.constants import HUGGINGFACE_MEDIA_TASKS

from owlsight.processors.base import TextGenerationProcessor
from owlsight.processors.text_generation_processors import TextGenerationProcessorTransformers
from owlsight.processors.constants import DEFAULT_MAX_TOKENS, DEFAULT_TEMPERATURE
from owlsight.utils.custom_classes import MediaObject
from owlsight.utils.logger import logger


class MediaPreprocessor:
    """
    Handles preprocessing for different media types and integrates with text generation.

    This class preprocesses media inputs (images, audio, documents) before passing them
    to the appropriate model pipeline.
    """

    def __init__(self, task: str):
        """
        Initialize preprocessor for specific task.

        Parameters
        ----------
        task : str
            The task to handle. Must be one of HUGGINGFACE_MEDIA_TASKS or a text task.
        """
        self.task = task
        self._validate_task()

    def _validate_task(self) -> None:
        """Validate that the task is supported."""
        if self.task not in HUGGINGFACE_MEDIA_TASKS and not self.task.endswith("generation"):
            raise ValueError(
                f"Task {self.task} is not supported. Must be one of {HUGGINGFACE_MEDIA_TASKS} "
                f"or end with 'generation'"
            )

    def preprocess_input(self, input_data: Union[str, bytes, Path], question: Optional[str] = None) -> Any:
        """
        Preprocess input data based on task type.

        Parameters
        ----------
        input_data : Union[str, bytes, Path]
            The input data. Can be a file path, URL, or bytes.
        question : Optional[str]
            Question for VQA or document QA tasks.

        Returns
        -------
        Dict[str, Any]
            Preprocessed data in format expected by the model.
        """
        if self.task not in HUGGINGFACE_MEDIA_TASKS:
            raise ValueError(
                f"Task {self.task} is not supported for media preprocessing. Should be one of {HUGGINGFACE_MEDIA_TASKS}"
            )

        try:
            if isinstance(input_data, (str, Path)):
                input_data = self._load_from_path_or_url(input_data)

            if self.task == "automatic-speech-recognition":
                return self._preprocess_audio(input_data)
            elif self.task in ["image-to-text", "visual-question-answering", "document-question-answering"]:
                processed = self._preprocess_image(input_data)
                if question and self.task in ["visual-question-answering", "document-question-answering"]:
                    return {"image": processed, "question": question}
            else:
                raise ValueError(f"Task {self.task} is not supported for media preprocessing.")
            return processed

        except Exception:
            logger.error(f"Error preprocessing input for task {self.task}: {traceback.format_exc()}")
            raise

    def _load_from_path_or_url(self, source: Union[str, Path]) -> bytes:
        """Load data from file path or URL."""
        if isinstance(source, str) and source.startswith(("http://", "https://")):
            response = requests.get(source, timeout=10)
            response.raise_for_status()
            return response.content
        else:
            p = Path(source)
            if not p.exists():
                raise FileNotFoundError(f"File not found: {source}")
            return p.read_bytes()

    def _preprocess_audio(self, audio_data: bytes) -> Dict[str, Any]:
        """Preprocess audio data."""
        # Convert to numpy array
        audio_array = np.frombuffer(audio_data, dtype=np.int16)
        audio_array = audio_array.astype(np.float32) / 32768.0

        # Convert stereo to mono if needed
        if len(audio_array.shape) > 1:
            audio_array = audio_array.mean(axis=1)

        return {"array": audio_array, "sampling_rate": 16000}  # Standard sampling rate for most models

    def _preprocess_image(self, image_data: bytes) -> Image.Image:
        """Preprocess image data."""
        image = Image.open(io.BytesIO(image_data))
        return image


class MultiModalProcessorTransformers(TextGenerationProcessor):
    def __init__(self, model_id: str, task: str, save_history: bool = False, system_prompt: str = "", **kwargs):
        if task not in HUGGINGFACE_MEDIA_TASKS:
            raise ValueError(
                f"Task {task} is not supported for media preprocessing. Should be one of {HUGGINGFACE_MEDIA_TASKS}"
            )

        super().__init__(model_id=model_id, save_history=save_history, system_prompt=system_prompt)
        self.task = task
        self.text_processor = TextGenerationProcessorTransformers(model_id=model_id, task=task, **kwargs)
        self.media_preprocessor = MediaPreprocessor(self.text_processor.task)

    def generate(
        self,
        input_data: str,
        media_objects: Dict[str, MediaObject],
        stopwords: Optional[List[str]] = None,
        max_new_tokens: int = DEFAULT_MAX_TOKENS,
        temperature: float = DEFAULT_TEMPERATURE,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> str:
        # First prepare the generation parameters
        input_data, generate_kwargs = self.text_processor.prepare_generation(
            input_data=input_data,
            max_new_tokens=max_new_tokens,
            temperature=temperature,
            stopwords=stopwords,
            streaming=False,
            generation_kwargs=generation_kwargs,
            apply_chat_template=False,
        )
        generate_kwargs.pop("eos_token_id", None)

        # Extract any referenced media objects and their positions if in the input text
        media_refs = re.finditer(r"__MEDIA_\d+__", input_data)

        # For each media reference, preprocess the media and store question if present
        preprocessed_data = []
        for ref in media_refs:
            media_id = ref.group()
            media_object = media_objects[media_id]

            # Get the question from the input text before the media reference
            text_before = input_data[: ref.start()].strip()
            question = text_before if text_before else None

            # Preprocess the media file
            preprocessed = self.media_preprocessor.preprocess_input(media_object.path, question)
            preprocessed_data.append(preprocessed)

        # If we have only one media object, unpack it
        if len(preprocessed_data) == 1:
            preprocessed_data = preprocessed_data[0]

        try:
            response = self.text_processor.pipe(preprocessed_data, generate_kwargs=generate_kwargs)
            response = str(response)
            print(response)
        except Exception as e:
            logger.error(f"Error generating text with media input: {traceback.format_exc()}")
            raise

        self.update_history(str(input_data), response.strip())
        return response

    def preprocess_input(self, input_data: Union[str, bytes, Path], question: Optional[str] = None) -> Any:
        processed = self.media_preprocessor.preprocess_input(input_data, question)
        return processed



def validate_key_is_nested_one_layer(key: str) -> bool:
    """
    Check if the given key is nested only one layer deep.

    A key is considered nested one layer deep if it contains at most one dot ('.') character.

    Parameters:
    ----------
        key (str): The key to validate.

    Returns:
        bool: True if the key is nested only one layer deep, False otherwise.
    """
    return key.count(".") <= 1



from typing import Any, Optional, Dict
import traceback
import pkgutil
import ast

from owlsight.processors.base import TextGenerationProcessor
from owlsight.processors.helper_functions import (
    select_processor_type,
)
from owlsight.ui.console import get_user_choice
from owlsight.configurations.config_manager import ConfigManager
from owlsight.rag.python_lib_search import search_python_libs
from owlsight.hugging_face.core import show_and_return_model_data
from owlsight.hugging_face.constants import HUGGINGFACE_MEDIA_TASKS
from owlsight.utils.helper_functions import convert_to_real_type
from owlsight.utils.deep_learning import free_memory
from owlsight.utils.constants import get_pickle_cache, DEFAULTS
from owlsight.utils.logger import logger

class TextGenerationManager:
    def __init__(self, config_manager: ConfigManager):
        """
        Manage the lifecycle of a TextGenerationProcessor and its interaction with the configuration during runtime of the CLI app.

        Parameters
        ----------
        processor : TextGenerationProcessor
            An instance of the processor (either Transformers or Onnx).
        config_manager : ConfigManager
            Configuration dictionary to manage settings for the processor.
        """
        self.config_manager = config_manager
        self.processor: Optional[TextGenerationProcessor] = None

    def generate(self, input_data: str, media_objects: Optional[Dict[str, dict]] = None) -> str:
        """
        Generate text using the processor.
        """
        task = self.config_manager.get("huggingface.task")
        kwargs = self.config_manager.get("generate", {})
        if media_objects or task in HUGGINGFACE_MEDIA_TASKS:
            generated_text = self.processor.generate(input_data, media_objects=media_objects, **kwargs)
        else:
            generated_text = self.processor.generate(input_data, **kwargs)
        if task in HUGGINGFACE_MEDIA_TASKS:
            try:
                result = ast.literal_eval(generated_text)
            except Exception:
                logger.error(f"Error evaluating generated text: {traceback.format_exc()}")
            if not result:
                logger.warning(f"No text generated for media task '{task}'.")
                logger.warning("Use the special [[]] syntax to pass media objects to the model.")
                for mediatype in ["image", "audio", "video"]:
                    logger.warning(f"For example: '[[{mediatype}:path/to/{mediatype}]]'")
                logger.warning("Or for QA: 'What color is the car? [[image:path/to/image.jpg]]'")

        return generated_text

    def update_config(self, key: str, value: Any):
        """
        Update the configuration dynamically. If 'model_id' is updated, reload the processor.
        """
        if key.endswith(".back"):
            return  # Do not set the "back" key
        try:
            value = convert_to_real_type(value)
            self.config_manager.set(key, value)
            logger.info(f"Configuration updated: {key} = {value}")
        except Exception:
            logger.error(f"Error updating configuration for key '{key}': {traceback.format_exc()}")
            return

        outer_key, inner_key = key.split(".", 1)
        if outer_key == "model":
            if inner_key == "model_id":
                self.load_model_processor(reload=self.processor is not None)
            else:
                if self.processor is None:
                    logger.error(
                        "Processor is not initialized yet. Assign a model_id first, to initialize a model for the processor."
                    )
                    return
                if hasattr(self.processor, inner_key):
                    setattr(self.processor, inner_key, value)
                    logger.info(f"Processor updated: {inner_key} = {value}")
                else:
                    logger.warning(f"'{inner_key}' not found in self.processor, meaning it was not updated")
                    logger.warning(
                        "It is possible that this value is only set during initialization of self.processor."
                    )
                    logger.warning("Consider loading the model from a config file to update this value.")
        elif outer_key == "rag":
            rag_is_active = self.config_manager.get("rag.active", False)
            if rag_is_active:
                library = self.config_manager.get("rag.target_library", "")
                if not library:
                    logger.error("No library provided. Please set a library in the configuration.")
                    return

                # get all libs without the _ prefix and in sorted order
                available_libraries = [
                    module.name for module in pkgutil.iter_modules() if not module.name.startswith("_")
                ]
                if library not in available_libraries:
                    logger.error(f"Library '{library}' not found in the current Python session.")
                    logger.error(f"available libraries: {sorted(available_libraries)}")
                    return
                elif inner_key == "search_query":
                    search_query = self.config_manager.get("rag.search_query", "")
                    if not search_query:
                        logger.error("No example prompt provided. Please set an example prompt in the configuration.")
                        return
                    top_k = self.config_manager.get("rag.top_k", DEFAULTS[outer_key]["top_k"])
                    context = search_python_libs(library, search_query, top_k, cache_dir=get_pickle_cache())
                    print(f"Context for library '{library}' with top_k={top_k}:\n{context}")
        elif outer_key == "huggingface":
            if inner_key == "search":
                # search models from huggingface
                model_search = self.config_manager.get("huggingface.search", DEFAULTS["huggingface"]["search"])
                top_k = self.config_manager.get("huggingface.top_k", DEFAULTS["huggingface"]["top_k"])
                task = self.config_manager.get("huggingface.task", DEFAULTS["huggingface"]["task"])
                model_dict = show_and_return_model_data(model_search, top_n_models=top_k, task=task)
                if not model_dict:
                    logger.error("No models found. Please try a different search query.")
                    return
                # set list of models from model_dict to select_model
                self.config_manager.set("huggingface.select_model", list(model_dict.keys()))
            elif inner_key == "select_model":
                select_model = self.config_manager.get(
                    "huggingface.select_model", DEFAULTS["huggingface"]["select_model"]
                )
                if not select_model:
                    logger.error("No model provided. Please set a model in the configuration.")
                    return
                if not isinstance(select_model, str):
                    logger.error("Model must be a string. Please set a model in the configuration.")
                    return
                # select and load a model from huggingface
                self.config_manager.set("model.model_id", select_model)
                task = self.config_manager.get("huggingface.task", DEFAULTS["huggingface"]["task"])
                exc = self.load_model_processor(reload=self.processor is not None)
                if exc:
                    if select_model.lower().endswith("gguf"):
                        gguf_list = str(exc).split("Available Files:")[1].strip()
                        if gguf_list:
                            gguf_list = [file for file in ast.literal_eval(gguf_list) if file.endswith("gguf")]
                            gguf_menu = {
                                "back": None,
                                "Choose a GGUF model": gguf_list,
                            }
                            gguf__filename = get_user_choice(gguf_menu)
                            if gguf__filename:
                                self.config_manager.set("model.gguf__filename", gguf__filename)
                                self.load_model_processor(reload=self.processor is not None)
                        else:
                            logger.warning("No gguf-list could be inferred")
            elif inner_key == "task":
                task = self.config_manager.get("huggingface.task", DEFAULTS["huggingface"]["task"])
                self.config_manager.set("huggingface.task", task)

    def save_config(self, path: str):
        """
        Save the configuration to a file.
        """
        self.config_manager.save(path)

    def load_config(self, path: str):
        """
        Load the configuration from a file.
        """
        loading_succesful = self.config_manager.load(path)
        if loading_succesful:
            self.load_model_processor(reload=self.processor is not None)

    # TODO: remove mechanism with **processor_kwargs and **kwargs in __init__ of TextGenerationProcessor classes
    # It is prone to errors.
    def load_model_processor(self, reload=False) -> None | Exception:
        """
        Load the model processor with a 'model_id', to load the correct model and tokenizer.

        Parameters
        ----------
        reload : bool, optional
            If True, reload the processor with the same model_id, by default False.
            Assumes that the processor is already initialized with another model.

        Returns
        -------
        None | Exception
            None if successful, otherwise an exception is returned.
        """
        model_kwargs = self.config_manager.get("model", {})
        task = self.config_manager.get("huggingface.task", DEFAULTS["huggingface"]["task"])
        processor_kwargs = {"task": task, **model_kwargs}

        model_id = self.config_manager.get("model.model_id", "")
        if not model_id:
            logger.error("No model_id provided. Please set a model_id in the configuration.")
            return

        logger.info(f"Loading processor with new model_id: {model_id}")
        processor_type = select_processor_type(model_id, task=task)

        try:
            if reload:
                if self.processor is None:
                    raise ValueError("Processor is not initialized yet. Cannot reload.")
                # Save the history from the old processor
                old_history = self.processor.history

                # Inmediately overwrite the processor with a new instance to save memory
                self.processor = None
                free_memory()

                self.processor = processor_type(**processor_kwargs)
                self.processor.history = old_history
            else:
                self.processor = processor_type(**processor_kwargs)
        except Exception as e:
            logger.error(f"Error loading model_processor: {traceback.format_exc()}")
            return e

        logger.info(f"Processor reloaded with model_id: {model_id}")

    def get_processor(self) -> TextGenerationProcessor:
        """
        Return the current processor instance.
        """
        return self.processor

    def get_config(self) -> dict:
        """
        Return the current configuration as dictionary.
        """
        return self.config_manager._config

    def get_config_choices(self) -> dict:
        """
        Return the available configuration choices.

        Returns
        -------
        dict
            Dictionary with the available configuration choices.
        """
        return self.config_manager.config_choices

    def get_config_key(self, key: str, default: Any = None) -> Any:
        """
        Get the value of a key in the configuration.
        """
        return self.config_manager.get(key, default)

    # def _add_media_objects_if_applicable(self, media_objects, kwargs) -> Dict[str, Any]:
    #     processor_generate = self.processor.generate
    #     signature = inspect.signature(processor_generate)
    #     if media_objects and "media_objects" in signature.parameters:
    #         kwargs["media_objects"] = media_objects

    #     return kwargs



from transformers.pipelines import SUPPORTED_TASKS

TASK_TO_AUTO_MODEL = {k: v["pt"][0] for k, v in SUPPORTED_TASKS.items()}

HUGGINGFACE_TASKS = [None] + [
    "text-generation",
    "text2text-generation",
    "translation",
    "summarization",
]

HUGGINGFACE_MEDIA_TASKS = [
    "image-to-text",
    "automatic-speech-recognition",
    "visual-question-answering",
    "document-question-answering",
]

# MEDIA_PIPELINE_MAPPING = {
#     "image-to-text": "image",
#     "automatic-speech-recognition": "audio",
#     "visual-question-answering": "image",
#     "document-question-answering": "image",
# }

HUGGINGFACE_TASKS.extend(HUGGINGFACE_MEDIA_TASKS)



import gc
import subprocess
import torch

from owlsight.utils.logger import logger




def free_memory():
    """Free up memory and reset stats."""
    gc.collect()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.reset_peak_memory_stats()


def print_memory_stats(device: torch.device):
    """Print two different measures of GPU memory usage."""
    print(
        f"Max memory allocated: {torch.cuda.max_memory_allocated(device) / 1e9:.2f} GB"
    )
    print(
        f"Max memory reserved: { torch.cuda.max_memory_reserved(device) / 1e9:.2f} GB"
    )


def calculate_model_size(model) -> float:
    param_size = 0
    for param in model.parameters():
        param_size += param.nelement() * param.element_size()
    buffer_size = 0
    for buffer in model.buffers():
        buffer_size += buffer.nelement() * buffer.element_size()
    size_all_mb = (param_size + buffer_size) / 1024**2
    return size_all_mb


def check_gpu_and_cuda():
    """Checks if a CUDA-capable GPU is available and if CUDA is installed."""
    if torch.cuda.is_available():
        gpu = torch.cuda.get_device_name(0)
        logger.info(f"GPU found: {gpu}")
        logger.info(
            "CUDA-capable GPU is available and PyTorch is built with CUDA support."
        )

    cuda_version = None
    try:
        output_cuda = subprocess.check_output(["nvcc", "--version"]).decode("utf-8")
        cuda_version = output_cuda[
            output_cuda.find("release")
            + len("release")
            + 1 : output_cuda.find(",", output_cuda.find("release"))
        ]
        logger.info("CUDA %s is installed.", cuda_version)
    except subprocess.CalledProcessError:
        logger.warning(
            "Warning: CUDA-capable GPU is available, but CUDA is not installed. Please install CUDA."
        )
    except Exception as e:
        logger.error("%s", e)

    if torch.cuda.is_available():
        logger.info("CUDA-capable GPU is available for PyTorch. You are all set!")
    else:
        logger.warning(
            "Cuda is currently unavailable. This could be expected if no GPU is available. If not, please visit 'https://pytorch.org/get-started/locally/' to install a compatible version.\nrun command 'pip uninstall torch torchvision torchaudio' and find run the right version of PyTorch for your CUDA version.",
        )


def log_reserved_memory():
    """Logs the reserved memory on the GPU and CPU."""
    if torch.cuda.is_available():
        gpu_reserved = torch.cuda.memory_reserved(0)
        gpu_free = torch.cuda.max_memory_allocated(0) - torch.cuda.memory_allocated(0)
        logger.info("GPU Memory - Reserved: %s, Free: %s", gpu_reserved, gpu_free)
    else:
        logger.info("CUDA not available. GPU memory stats cannot be logged.")

    try:
        cpu_stats = torch.cuda.memory_stats()
        cpu_reserved = cpu_stats.get("reserved_host_bytes.all.current", "Not available")
    except AttributeError:
        cpu_reserved = "Not available due to PyTorch version or configuration."

    logger.info("CPU Memory - Reserved: %s", cpu_reserved)


def bfloat16_is_supported():
    """
    Check if the current GPU supports bfloat16 data type using PyTorch.

    Returns:
        bool: True if bfloat16 is supported, False otherwise.
    """
    if not torch.cuda.is_available():
        return False

    try:
        _ = torch.tensor([1.0, 2.0], dtype=torch.bfloat16, device="cuda")
        return True
    except Exception:
        return False


def calculate_memory_for_model(n_bilion_parameters: int, n_bit: int = 32) -> float:
    """
    Calculate the memory required for a model in GB.

    Parameters:
    n_bilion_parameters (int): The number of parameters in the model in billions.
    n_bit (int): The number of bits used to represent the model parameters. Default is 32. Quantized models use 16/8/4 bits.
    """
    return ((n_bilion_parameters * 4) / (32 / n_bit)) * 1.2


def calculate_available_vram() -> float:
    """
    Calculate the available VRAM on the GPU in GB.
    """
    if torch.cuda.is_available():
        available_memory = torch.cuda.get_device_properties(
            0
        ).total_memory - torch.cuda.memory_allocated(0)
        available_memory_gb = available_memory / 1024**3
        return available_memory_gb
    else:
        logger.warning("CUDA not available. Cannot calculate available VRAM.")
        return 0.0


def calculate_max_parameters_per_dtype():
    """
    Calculate the maximum number of parameters that can be run on the GPU
    for different data types (32-bit, 16-bit, 8-bit, 4-bit).
    """
    available_vram = calculate_available_vram()
    if available_vram > 0:
        logger.info(f"Available VRAM: {available_vram:.2f} GB")

        for bits in [32, 16, 8, 4]:
            max_params = available_vram / calculate_memory_for_model(
                1, bits
            )  # for 1 billion parameters
            logger.info(
                f"Maximum number of billion parameters for {bits}-bit model: {max_params:.2f} billion"
            )
    else:
        logger.warning("No available VRAM to calculate parameters.")


def get_best_device() -> str:
    """
    Check for best device and return the device name.
    """
    if torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    else:
        return "cpu"



import os
import sys
from typing import Any, List, Union
import venv
from contextlib import contextmanager
import subprocess
import sysconfig
import tempfile
from pathlib import Path

from owlsight.utils.helper_functions import os_is_windows, force_delete
from owlsight.utils.logger import logger


@contextmanager
def create_venv(pyenv_path: str) -> str:
    """
    Context manager to create and manage a Python virtual environment.

    Parameters
    ----------
    pyenv_path : str
        The path where the virtual environment will be created.

    Yields
    ------
    str
        Path to the pip executable within the created virtual environment.
    """
    venv.create(pyenv_path, with_pip=True)
    pip_path = os.path.join(pyenv_path, "Scripts" if os_is_windows() else "bin", "pip")
    yield pip_path


def in_venv() -> bool:
    """
    Check if the current Python process is running inside a virtual environment.

    Returns
    -------
    bool
    True if the current process is running inside a virtual environment, False otherwise.
    """
    return hasattr(sys, "real_prefix") or (hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix)


def get_lib_path(pyenv_path: str) -> str:
    """
    Get the path to the lib directory within the virtual environment.

    Parameters
    ----------
    pyenv_path : str
        The path to the (virtual) python environment.

    Returns
    -------
    str
        The path to the lib directory.
    """
    # Get the name of the site-packages directory
    site_packages = sysconfig.get_path("purelib", vars={"base": pyenv_path})
    return site_packages


def get_python_executable(pyenv_path: str) -> str:
    """
    Get the path to the Python executable within the virtual environment.

    Parameters
    ----------
    pyenv_path : str
        The path to the virtual environment.

    Returns
    -------
    str
        The path to the Python executable.
    """
    return os.path.join(pyenv_path, "Scripts" if os_is_windows() else "bin", "python")


def get_pyenv_path() -> str:
    """
    Get the path to the current (virtual) python environment.

    Returns
    -------
    bool
        The path to the current (virtual) python environment.
    """
    # if not in_venv():
    #     raise RuntimeError("Not running inside a virtual environment.")
    return sys.prefix


def get_pip_path(pyenv_path: str) -> str:
    """
    Get the path to the pip executable within the (virtual) python environment.

    Parameters
    ----------
    pyenv_path : str
        The path to the (virtual) python environment.

    Returns
    -------
    str
        The path to the pip executable.
    """
    return os.path.join(pyenv_path, "Scripts" if os_is_windows() else "bin", "pip")


def get_temp_dir(suffix: str) -> str:
    """
    Get an appropriate temporary directory path that the user has write permissions for.

    Parameters
    ----------
    suffix : str
        The suffix to be appended to the temporary directory path.
        e.g., ".owlsight_temp"

    Returns
    -------
    str
        The path to a writable temporary directory
    """
    # Try user's home directory first
    user_temp = Path.home() / suffix
    if user_temp.exists():
        # remove the directory forcefully
        try:
            os.rmdir(user_temp)
        except Exception:
            force_delete(user_temp)
    try:
        os.makedirs(user_temp, exist_ok=True)
        return user_temp
    except Exception:
        # Fall back to system temp directory
        return tempfile.gettempdir()


def install_python_modules(module_names: Union[str, List[str]], pip_path: str, target_dir: str, *args: Any) -> bool:
    """
    Install one or more Python modules using pip into a specified directory and add it to sys.path.

    Parameters
    ----------
    module_names : Union[str, List[str]]
        The name of the module(s) to install. Can be a single module as a string or a list of modules.
    pip_path : str
        The path to the pip executable.
    target_dir : str
        The directory where the module(s) should be installed.
    *args : Any
        Additional arguments to pass to the pip install command (e.g., --extra-index-url).

    Returns
    -------
    bool
        True if all installations are successful, False otherwise.

    Examples
    --------
    >>> install_python_modules("some-package", pip_path, temp_dir, "--extra-index-url", "https://private-repo.com/simple")
    >>> install_python_modules(["some-package", "another-package"], pip_path, temp_dir)
    """

    # Convert module_name to a list if it's a string (comma-separated or space-separated)
    if isinstance(module_names, str):
        module_names = [name.strip() for name in module_names.split(" ")]

    success = True

    for module in module_names:
        pip_command = [pip_path, "install", "--target", target_dir, module] + list(args)
        try:
            # Install the module
            subprocess.check_call(pip_command)
            logger.info(f"Successfully installed {module} into {target_dir}")

            # Add target_dir to sys.path so that installed modules can be imported
            if target_dir not in sys.path:
                sys.path.insert(0, target_dir)

        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to install {module}. Error: {e}")
            success = False

    return success



import os
import sys
import logging

from pathlib import Path
import subprocess
import pytest
from typing import Tuple
from transformers import AutoTokenizer

sys.path.append("src")
from owlsight.processors.text_generation_processors import TextGenerationProcessorOnnx

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def download_model(repo_url: str, destination: Path) -> bool:
    """
    Download model from HuggingFace using git, with proper error handling.
    
    Args:
        repo_url: HuggingFace repository URL
        destination: Local destination path
    
    Returns:
        bool: True if successful, False otherwise
    """
    try:
        if destination.exists():
            logger.info("Model directory already exists, skipping download")
            return True
            
        logger.info(f"Downloading model from {repo_url}")
        result = subprocess.run(
            ["git", "clone", repo_url, str(destination)],
            capture_output=True,
            text=True,
            check=True
        )
        logger.info("Model downloaded successfully")
        return True
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to clone repository: {e.stderr}")
        return False
    except Exception as e:
        logger.error(f"Unexpected error during model download: {e}")
        return False

@pytest.fixture
def setup_gguf_processor() -> Tuple[TextGenerationProcessorOnnx, AutoTokenizer]:
    """
    Fixture to set up the GGUF processor and tokenizer with proper error handling.
    
    Returns:
        Tuple[TextGenerationProcessorOnnx, AutoTokenizer]: Processor and tokenizer
    """
    model_repo = "https://huggingface.co/llmware/tiny-llama-chat-onnx"
    destination_dir = Path("tests/data/tiny-llama-chat-onnx")
    tokenizer_id = "TinyLlama/TinyLlama-1.1B-Chat-v1.0"

    # Ensure parent directory exists
    destination_dir.parent.mkdir(parents=True, exist_ok=True)

    # Download model if needed
    if not download_model(model_repo, destination_dir):
        pytest.skip("Failed to download model")

    try:
        logger.info("Initializing processor and tokenizer...")
        processor = TextGenerationProcessorOnnx(
            model_id=str(destination_dir),
            onnx__tokenizer=tokenizer_id,
        )
        logger.info("Processor initialized successfully")
        return processor, processor.transformers_tokenizer
    except Exception as e:
        logger.error(f"Failed to initialize processor: {e}")
        pytest.skip(f"Processor initialization failed: {e}")

@pytest.mark.slow  # Mark as slow test
def test_gguf_generate_response(setup_gguf_processor):
    """Test that the GGUF processor generates a valid response."""
    processor, _ = setup_gguf_processor
    prompt = "test prompt"
    max_new_tokens = 128

    logger.info("Generating response...")
    response = processor.generate(prompt, max_new_tokens=max_new_tokens)
    logger.info("Response generated successfully")
    
    assert isinstance(response, str), "Generated response should be a string."
    assert len(response) > 0, "Generated response should not be empty."

@pytest.mark.slow
def test_gguf_token_count_within_tolerance(setup_gguf_processor):
    """Test that the GGUF generated token count is within the acceptable range."""
    processor, tokenizer = setup_gguf_processor
    prompt = "test prompt"
    max_new_tokens = 128
    tolerance_fraction = 0.5

    logger.info("Testing token count...")
    response = processor.generate(prompt, max_new_tokens=max_new_tokens)

    response_tokens = tokenizer.tokenize(response)
    prompt_tokens = tokenizer.tokenize(prompt)
    new_tokens = response_tokens[len(prompt_tokens):]

    lower_bound = max_new_tokens - (max_new_tokens * tolerance_fraction)
    upper_bound = max_new_tokens + (max_new_tokens * tolerance_fraction)

    logger.info(f"New tokens generated: {len(new_tokens)}")
    logger.info(f"Expected range: {lower_bound:.2f} - {upper_bound:.2f}")

    assert lower_bound <= len(new_tokens) <= upper_bound, (
        f"Expected approximately {max_new_tokens} tokens "
        f"(range: {lower_bound:.2f} - {upper_bound:.2f}), got {len(new_tokens)}"
    )

@pytest.mark.slow
def test_gguf_prompt_tokens_exclusion(setup_gguf_processor):
    """Test that the GGUF response excludes prompt tokens when counting new tokens."""
    processor, tokenizer = setup_gguf_processor
    prompt = "test prompt"
    max_new_tokens = 128

    logger.info("Testing prompt token exclusion...")
    response = processor.generate(prompt, max_new_tokens=max_new_tokens)

    response_tokens = tokenizer.tokenize(response)
    prompt_tokens = tokenizer.tokenize(prompt)

    logger.info(f"Response tokens: {len(response_tokens)}")
    logger.info(f"Prompt tokens: {len(prompt_tokens)}")

    assert len(response_tokens) > len(prompt_tokens), "Response tokens should exceed prompt tokens."

if __name__ == "__main__":
    pytest.main([__file__, "-v"])


import re
import inspect
import traceback
from typing import Dict, List, Optional
from types import MethodType

from owlsight.utils.logger import logger


class TransformersArgumentInferer:
    """
    Infer arguments of a method/function from its docstring or signature.
    This assumes the docstring in the style of the HuggingFace Transformers library.
    """

    def __init__(self):
        self._last_returned_value: Optional[Dict[str, dict | str]] = None

    def transform_inferred_arguments_to_text(self, inferred_arguments: Dict[str, dict | str]) -> Optional[str]:
        """
        Transform the dict acquired from self.__call__ to a string.\n
        For example, as placeholder text in an UI application."""
        if not isinstance(inferred_arguments, dict):
            raise TypeError(f"Expected dict, got {type(inferred_arguments)} instead.")

        if not inferred_arguments:
            raise ValueError("No arguments found.")

        if all(isinstance(value, str) for value in inferred_arguments.values()):
            return "\n".join(val for _, val in inferred_arguments.items())

        return "\n".join([f"{arg}: {arg_info['types']}" for arg, arg_info in inferred_arguments.items()])

    def __call__(self, func: MethodType, docstring: Optional[str] = None) -> Dict[str, dict | str]:
        """
        Infer the arguments of the specified function or method.

        Parameters
        ----------
        func : Callable
            The method or function
        docstring : str, optional
            The docstring of the method, by default None
            In some cases, the docstring is defined outside the func, so it can be passed here.

        Returns
        -------
        Dict[str, dict] | List[str]
            A dictionary with the argument names as keys and a dictionary with the argument details as values,
            or a list of obliged argument names.
        """
        if not callable(func):
            raise TypeError(f"Expected a callable (function or method), got {type(func)} instead.")
        if docstring is None:
            docstring = func.__doc__

        if not docstring:
            logger.error("Docstring not found for func %s  Using inspect to infer arguments.", func.__name__)
            signature = inspect.signature(func)
            arguments: List[str] = self._infer_arguments_from_signature(signature)
            return arguments

        try:
            arguments: dict = self._infer_arguments_from_doc_str(docstring)
        except Exception:
            logger.error(
                "Failed to infer arguments from docstring due to:\n%s",
                traceback.format_exc(),
            )
            logger.error("Using inspect to infer arguments.")
            signature = inspect.signature(func)
            arguments = self._infer_arguments_from_signature(signature)

        self._last_returned_value = arguments

        return arguments

    def _infer_arguments_from_doc_str(self, docstring: str) -> dict:
        arg_pattern = re.compile(r"\s*(\w+)\s+\(([^)]+)\):\s*(.*)")

        lines = docstring.strip().split("\n")
        arg_lines = []
        current_arg = None

        for line in lines:
            if re.match(arg_pattern, line):
                if current_arg:
                    arg_lines.append(current_arg)
                current_arg = line
            elif current_arg:
                current_arg += " " + line.strip()

        if current_arg:
            arg_lines.append(current_arg)

        arguments = {}
        for line in arg_lines:
            match = re.match(arg_pattern, line)
            if match:
                arg_name, arg_types, description = match.groups()
                if arg_name in ["args", "kwargs"]:
                    continue
                is_optional = "optional" in arg_types.lower()
                arguments[arg_name] = {
                    "types": arg_types,
                    "optional": is_optional,
                    "description": description,
                }

        return arguments

    def _infer_arguments_from_signature(self, signature: inspect.Signature) -> Dict[str, str]:
        return {name: str(param) for name, param in signature.parameters.items() if name not in "self"}



from typing import Optional, List, Dict, Any, Union, Tuple, Generator
import os
import time
import traceback
import threading
from ast import literal_eval
from functools import lru_cache

import torch
from transformers import (
    BitsAndBytesConfig,
    TextIteratorStreamer,
    AutoTokenizer,
    AutoModel,
    PreTrainedTokenizer,
    pipeline,
    Pipeline,
)
from owlsight.processors.base import TextGenerationProcessor
from owlsight.processors.constants import (
    DEFAULT_TASK,
    GENERATION_THREAD_TIMEOUT,
    DEFAULT_MAX_TOKENS,
    DEFAULT_TEMPERATURE,
)
from owlsight.utils.threads import ThreadNotKilledError
from owlsight.utils.custom_exceptions import QuantizationNotSupportedError, InvalidGGUFFileError
from owlsight.utils.custom_classes import StopWordCriteria
from owlsight.utils.deep_learning import get_best_device, bfloat16_is_supported
from owlsight.hugging_face.constants import SUPPORTED_TASKS
from owlsight.utils.helper_functions import check_invalid_input_parameters
from owlsight.utils.logger import logger

ONNX_MSG = "ONNX Runtime is disabled. Use 'pip install owlsight[onnx]' or install [onnxruntime-genai, onnxruntime-genai-cuda] seperately"

try:
    import onnxruntime_genai as og
except ImportError:
    logger.warning("Support for ONNX models is disabled.")
    og = None

try:
    from llama_cpp import Llama
except ImportError:
    logger.warning(
        "Support for GGUF models is disabled, because llama-cpp is not found. Install it using 'pip install llama-cpp-python'."
    )
    Llama = None


class TextGenerationProcessorTransformers(TextGenerationProcessor):
    def __init__(
        self,
        model_id: str,
        transformers__device: Optional[str] = None,
        transformers__quantization_bits: Optional[int] = None,
        transformers__stream: bool = True,
        transformers__use_fp16: bool = False,
        bnb_kwargs: Optional[dict] = None,
        tokenizer_kwargs: Optional[dict] = None,
        model_kwargs: Optional[dict] = None,
        task: Optional[str] = None,
        save_history: bool = False,
        system_prompt: str = "",
        **kwargs,
    ):
        """
        Text generation processor using transformers library.

        Parameters
        ----------
        model_id : str
            The model ID to use for generation.
            Usually the name of the model or the path to the model.
        transformers__device : str
            The device to use for generation. Default is None, where the best available device is checked out of the possible devices.
        transformers__quantization_bits : Optional[int]
            The number of quantization bits to use for the model. Default is None.
        transformers__stream : bool
            Whether to use streaming generation. Default is True.
        transformers__use_fp16 : bool
            Whether to use FP16 for the model. This will not work for cpu, as FP16 is not supported on CPU.
            Checks if bfloat16 is supported and will use this if available, else uses torch.float16.
        bnb_kwargs : Optional[dict]
            Additional keyword arguments for BitsAndBytesConfig. Default is None.
        tokenizer_kwargs : Optional[dict]
            Additional keyword arguments for the tokenizer. Default is None.
        model_kwargs : Optional[dict]
            Additional keyword arguments for the model. Default is None.
        task : Optional[str]
            The task to use for the pipeline. Default is None, where the task is set to "text-generation".
        save_history : bool
            Set to True if you want model to generate responses based on previous inputs.
        system_prompt : str
            The system prompt to prepend to the input text.
        """
        if task and task not in SUPPORTED_TASKS:
            raise ValueError(f"Task '{task}' is not supported. Supported tasks are: {list(SUPPORTED_TASKS.keys())}")

        super().__init__(model_id, save_history, system_prompt)

        # Initialize configuration
        self.transformers__device = transformers__device or get_best_device()
        self.transformers__stream = transformers__stream
        self.transformers__quantization_bits = transformers__quantization_bits
        self.transformers__use_fp16 = transformers__use_fp16
        self.bnb_kwargs = bnb_kwargs or {}
        self.tokenizer_kwargs = tokenizer_kwargs or {}
        self.model_kwargs = model_kwargs or {}
        self.task = task or DEFAULT_TASK

        # Set device and dtype configuration
        self._torch_dtype = self._determine_torch_dtype()

        # Initialize model components
        self._setup_tokenizer_and_model_kwargs()
        self.pipe = self._setup_pipeline()
        self.streamer = self._setup_streamer() if self.transformers__stream else None

    def _determine_torch_dtype(self) -> Any:
        """Determine appropriate torch dtype based on configuration."""
        if self.transformers__use_fp16:
            if self.transformers__device == "cpu":
                raise TypeError("FP16 is not supported on CPU.")
            return self._get_correct_fp16_dtype()
        return torch.float32 if self.transformers__device == "cpu" else "auto"

    @lru_cache(maxsize=1)
    def _get_correct_fp16_dtype(self) -> torch.dtype:
        """Get correct FP16 dtype based on hardware support."""
        return torch.bfloat16 if bfloat16_is_supported() else torch.float16

    def _setup_tokenizer_and_model_kwargs(self) -> Tuple[AutoTokenizer, AutoModel]:
        """Load and configure tokenizer and model."""
        if self.transformers__quantization_bits and self.transformers__device in ["cpu", "mps"]:
            raise QuantizationNotSupportedError("Quantization not supported on CPU or MPS.")

        quantization_config = self._get_quantization_config()
        self.model_kwargs = self._prepare_model_kwargs(quantization_config)

        self.tokenizer = None
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.model_id, trust_remote_code=True, **self.tokenizer_kwargs
            )
        except Exception:
            logger.error(f"Failed to load tokenizer for model {self.model_id}: {traceback.format_exc()}")

    def _flash_attention_is_available(self) -> bool:
        """Check if flash attention is available."""
        try:
            from flash_attn import flash_attn_fn

            return True
        except ImportError:
            return False

    def _get_quantization_config(self) -> Optional[BitsAndBytesConfig]:
        """Get quantization configuration if applicable."""
        if self.transformers__quantization_bits == 4:
            return BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_compute_dtype=self._get_correct_fp16_dtype(),
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type="nf4",
                **self.bnb_kwargs,
            )
        elif self.transformers__quantization_bits == 8:
            return BitsAndBytesConfig(load_in_8bit=True, **self.bnb_kwargs)
        return None

    def _prepare_model_kwargs(self, quantization_config: Optional[BitsAndBytesConfig]) -> Dict[str, Any]:
        """Prepare model initialization kwargs."""
        kwargs = {
            "torch_dtype": self._torch_dtype,
            "quantization_config": quantization_config,
            "_attn_implementation": "flash" if self._flash_attention_is_available() else "eager",
        }
        kwargs.update(self.model_kwargs)
        return kwargs

    def _setup_pipeline(self) -> Pipeline:
        """Set up the generation pipeline using EAFP pattern.
        
        Attempts to create pipeline with device specification first.
        If that fails due to Accelerate, creates pipeline without device parameter.
        """
        pipeline_kwargs = {
            "task": self.task,
            "model": self.model_id,
            "tokenizer": self.tokenizer,
            "trust_remote_code": True,
            "model_kwargs": self.model_kwargs,
            "device": self.transformers__device  # Try with device first
        }
        
        try:
            return pipeline(**pipeline_kwargs)
        except ValueError as e:
            if "model has been loaded with `accelerate`" in str(e):
                # Remove device parameter and retry if using accelerate
                del pipeline_kwargs["device"]
                return pipeline(**pipeline_kwargs)
            raise  # Re-raise if it's a different ValueError

    def _setup_streamer(self) -> TextIteratorStreamer:
        """Set up text streaming if enabled."""
        return TextIteratorStreamer(
            self.pipe.tokenizer,
            skip_prompt=True,
            skip_special_tokens=True,
        )

    @torch.inference_mode()
    def generate(
        self,
        input_data: str,
        max_new_tokens: int = DEFAULT_MAX_TOKENS,
        temperature: float = DEFAULT_TEMPERATURE,
        stopwords: Optional[List[str]] = None,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Generate text response."""
        if self.transformers__stream:
            response = ""
            for text_chunk in self.generate_stream(
                input_data, max_new_tokens, temperature, stopwords, generation_kwargs
            ):
                print(text_chunk, end="", flush=True)
                response += text_chunk
            print()  # Print newline after generation is done
            return response
        return self._generate_non_stream(input_data, max_new_tokens, temperature, stopwords, generation_kwargs)

    @torch.inference_mode()
    def generate_stream(
        self,
        input_data: str,
        max_new_tokens: int = DEFAULT_MAX_TOKENS,
        temperature: float = DEFAULT_TEMPERATURE,
        stopwords: Optional[List[str]] = None,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Generator[str, None, None]:
        """Generate streaming text response."""
        if not self.transformers__stream:
            raise ValueError("Streaming is disabled. Enable with transformers__stream=True.")

        yield from self._generate_stream(input_data, max_new_tokens, temperature, stopwords, generation_kwargs)

    def prepare_generation(
        self,
        input_data: str,
        max_new_tokens: int,
        temperature: float,
        stopwords: Optional[List[str]],
        generation_kwargs: Optional[Dict[str, Any]],
        streaming: bool = False,
        apply_chat_template: bool = True,
    ) -> Tuple[str, Dict[str, Any]]:
        """
        Prepare generation parameters.

        Returns:
        --------
        Tuple[str, Dict[str, Any]]
            The input text and generation_kwargs.
        """
        if apply_chat_template:
            input_data = self.apply_chat_template(input_data, self.pipe.tokenizer)

        gen_kwargs = {
            "max_new_tokens": max_new_tokens,
            "eos_token_id": self.pipe.tokenizer.eos_token_id,
            "temperature": temperature if temperature > 0.0 else None,
            "do_sample": temperature > 0.0,
        }

        if stopwords:
            gen_kwargs["stopping_criteria"] = StopWordCriteria(
                prompts=[input_data],
                stop_words=stopwords,
                tokenizer=self.pipe.tokenizer,
            )

        if generation_kwargs:
            gen_kwargs.update(generation_kwargs)

        if streaming:
            gen_kwargs["streamer"] = self.streamer
            # gen_kwargs["num_beams"] = 1  # Required for streaming

        return input_data, gen_kwargs

    def _generate_non_stream(
        self,
        input_data: str,
        max_new_tokens: int,
        temperature: float,
        stopwords: Optional[List[str]],
        generation_kwargs: Optional[Dict[str, Any]],
    ) -> str:
        """Generate text without streaming."""
        templated_text, gen_kwargs = self.prepare_generation(
            input_data, max_new_tokens, temperature, stopwords, generation_kwargs
        )
        output = self.pipe(templated_text, **gen_kwargs)
        generated_text = output[0]["generated_text"][len(templated_text) :].strip()
        self.update_history(input_data, generated_text)
        return generated_text

    def _generate_stream(
        self,
        input_data: str,
        max_new_tokens: int,
        temperature: float,
        stopwords: Optional[List[str]],
        generation_kwargs: Optional[Dict[str, Any]],
    ) -> Generator[str, None, None]:
        """Generate streaming text."""
        templated_text, gen_kwargs = self.prepare_generation(
            input_data, max_new_tokens, temperature, stopwords, generation_kwargs, streaming=True
        )

        stop_event = threading.Event()
        generation_thread = threading.Thread(
            target=self._run_generation_thread,
            args=(templated_text, gen_kwargs, stop_event),
        )
        generation_thread.start()

        try:
            yield from self._stream_generator(stop_event, input_data)
        except Exception as e:
            logger.error(f"Streaming error: {traceback.format_exc()}")
            raise e
        finally:
            stop_event.set()
            generation_thread.join(timeout=GENERATION_THREAD_TIMEOUT)

            if generation_thread.is_alive():
                raise ThreadNotKilledError("Generation thread wasn't killed in time.")

    def _stream_generator(self, stop_event: threading.Event, input_data: str) -> Generator[str, None, None]:
        """Handle text stream generation."""
        generated_text = ""
        try:
            while not stop_event.is_set():
                try:
                    new_text = next(self.streamer)
                    generated_text += new_text
                    yield new_text
                except StopIteration:
                    break
                # Check for error after each iteration
                if hasattr(self.streamer, "error") and self.streamer.error is not None:
                    raise self.streamer.error
        finally:
            if generated_text:
                self.update_history(input_data, generated_text.strip())

    def _run_generation_thread(
        self, templated_text: str, gen_kwargs: Dict[str, Any], stop_event: threading.Event
    ) -> None:
        """Run generation in a separate thread."""
        try:
            self.pipe(templated_text, **gen_kwargs)
        except Exception as e:
            self.streamer.error = e  # Store error in streamer
            self.streamer.end()
        finally:
            stop_event.set()


class TextGenerationProcessorOnnx(TextGenerationProcessor):
    def __init__(
        self,
        model_id: str,
        onnx__tokenizer: Union[str, PreTrainedTokenizer],
        onnx__verbose: bool = False,
        onnx__num_threads: int = 1,
        save_history: bool = False,
        system_prompt: str = None,
        **kwargs,
    ):
        """
        Text generation processor using ONNX Runtime.

        Parameters
        ----------
        model_id : str
            The model ID to use for generation.
            Usually the name of the model or the path to the model.
        onnx__tokenizer : Union[str, PreTrainedTokenizer]
            The tokenizer to use for generation.
            If str, it should be the model ID of the tokenizer.
            else, it should be a PreTrainedTokenizer object.
            This tokenizer allows universal use of chat templates.
        onnx__verbose : bool
            Whether to print verbose logs.
        onnx__num_threads : int
            Number of threads to use for generation.
        save_history : bool
            Set to True if you want model to generate responses based on previous inputs.
        system_prompt : str
            The system prompt to prepend to the input text.
        """
        self._validate_model_tokenizer(model_id, onnx__tokenizer)

        super().__init__(model_id, save_history, system_prompt)
        self.onnx__verbose = onnx__verbose
        self.onnx__num_threads = onnx__num_threads

        self._set_tokenizer(onnx__tokenizer)
        self._set_environment_variables()
        self._initialize_model()

    def generate(
        self,
        input_data: str,
        max_new_tokens: int = 512,
        temperature: float = 0.0,
        stopwords: Optional[List[str]] = None,
        buffer_wordsize: int = 10,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> str:
        """
        Generate text using the ONNX model.

        Parameters
        ----------
        input_data : str
            The input text to generate a response for.
        max_new_tokens : int
            The maximum number of tokens to generate.
        temperature : float
            The temperature for sampling.
        stopwords : List[str], optional
            List of stop words to stop generation at.
        buffer_wordsize : int
            The buffer word size for generation.
            Larger buffer sizes will check later for stop words.
        generation_kwargs : Dict[str, Any], optional
            Additional keyword arguments for generation.
            Example: {"top_k": 50, "top_p": 0.95}
        """
        generator = self._prepare_generate(input_data, max_new_tokens, temperature, generation_kwargs)

        logger.info("Running generation loop ...")
        generated_text, buffer = "", ""
        token_counter = 0
        start = time.time()

        try:
            while not generator.is_done():
                generator.compute_logits()
                generator.generate_next_token()
                new_text = self.tokenizer_stream.decode(generator.get_next_tokens()[0])
                buffer += new_text
                token_counter += 1
                print(new_text, end="", flush=True)

                if len(buffer.split()) > buffer_wordsize:
                    generated_text += buffer
                    buffer = ""

                    if stopwords and any(stop_word in generated_text for stop_word in stopwords):
                        break

        except KeyboardInterrupt:
            logger.warning("Control+C pressed, aborting generation")

        generated_text += buffer
        del generator

        total_time = time.time() - start
        logger.info(f"Generation took {total_time:.2f} seconds")
        logger.info(f"Tokens per second: {token_counter / total_time:.2f}")

        self.update_history(input_data, generated_text.strip())

        return generated_text.strip()

    def generate_stream(
        self,
        input_data: str,
        max_new_tokens: int = 512,
        temperature: float = 0.0,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """
        Generate text using the ONNX model.

        Parameters
        ----------
        input_data : str
            The input text to generate a response for.
        max_new_tokens : int
            The maximum number of tokens to generate.
        temperature : float
            The temperature for sampling.
        generation_kwargs : Dict[str, Any], optional
            Additional keyword arguments for generation.
            Example: {"top_k": 50, "top_p": 0.95}
        """
        generator = self._prepare_generate(input_data, max_new_tokens, temperature, generation_kwargs)

        logger.info("Running generation loop ...")
        generated_text = ""

        try:
            while not generator.is_done():
                generator.compute_logits()
                generator.generate_next_token()
                new_text = self.tokenizer_stream.decode(generator.get_next_tokens()[0])
                generated_text += new_text
                yield new_text

        except KeyboardInterrupt:
            logger.warning("Control+C pressed, aborting generation")

        del generator

        self.update_history(input_data, generated_text.strip())

    def _prepare_generate(self, input_data, max_new_tokens, temperature, generation_kwargs):
        templated_text = self.apply_chat_template(input_data, self.transformers_tokenizer)

        search_options = {
            "max_length": max_new_tokens,
            "temperature": temperature,
            **(generation_kwargs or {}),
        }

        input_tokens = self.tokenizer.encode(templated_text)

        params = og.GeneratorParams(self.model)
        params.set_search_options(**search_options)
        params.input_ids = input_tokens
        generator = og.Generator(self.model, params)

        return generator

    def _set_tokenizer(self, onnx__tokenizer):
        if isinstance(onnx__tokenizer, str):
            self.transformers_tokenizer = AutoTokenizer.from_pretrained(onnx__tokenizer)
        else:
            self.transformers_tokenizer = onnx__tokenizer

    def _set_environment_variables(self) -> None:
        os.environ.update(
            {
                "OMP_NUM_THREADS": str(self.onnx__num_threads),
                "OMP_WAIT_POLICY": "ACTIVE",
                "OMP_SCHEDULE": "STATIC",
                "ONNXRUNTIME_INTRA_OP_NUM_THREADS": str(self.onnx__num_threads),
                "ONNXRUNTIME_INTER_OP_NUM_THREADS": "1",
            }
        )

    def _initialize_model(self) -> None:
        logger.info("Loading model...")
        self.model = og.Model(self.model_id)
        self.tokenizer = og.Tokenizer(self.model)
        self.tokenizer_stream = self.tokenizer.create_stream()
        logger.info(f"Model loaded using {self.onnx__num_threads} threads")
        logger.info("Tokenizer created")

    def _validate_model_tokenizer(self, model_id, onnx__tokenizer):
        if og is None:
            raise ImportError(ONNX_MSG)

        if not os.path.exists(model_id):
            raise FileNotFoundError(f"{model_id} does not exist! Ensure the model path is an existing local directory.")

        if not os.path.isdir(model_id):
            raise NotADirectoryError(f"{model_id} is not a directory! Ensure the model path is a directory.")

        model_path_contents = os.listdir(model_id)
        if not "genai_config.json" in model_path_contents:
            raise FileNotFoundError(
                f"{model_id} does not contain a genai_config.json! This file is required for ONNX models."
            )

        if not onnx__tokenizer:
            raise ValueError(
                "No tokenizer found! "
                "A tokenizer from the transformers library is required "
                "for ONNX models, to standardize chat templates."
                "Look into HuggingFace (https://huggingface.co) and find the fitting model to use."
            )


class TextGenerationProcessorGGUF(TextGenerationProcessor):
    def __init__(
        self,
        model_id: str,
        gguf__filename: str = "",
        gguf__verbose: bool = False,
        gguf__n_ctx: int = 512,
        gguf__n_gpu_layers: int = 0,
        gguf__n_batch: int = 512,
        gguf__n_cpu_threads: int = 1,
        save_history: bool = False,
        system_prompt: str = "",
        model_kwargs: Dict[str, Any] = None,
        **kwargs,
    ):
        """
        Text generation processor using GGUF models. Uses llama-cpp.Llama class under the hood.

        Parameters
        ----------
        model_id : str
            The model ID to use for generation.
            Usually the name of the model (on HuggingFace) or the path to the model.
        gguf__filename : str
            The filename of the model to load. This is required when loading a model from huggingface.
        gguf__verbose : bool
            Whether to print verbose logs from llama_cpp.LLama class.
        gguf__n_ctx : int
            The context size for the model.
        gguf__n_gpu_layers : int
            The number of layers to offload to the GPU.
        gguf__n_batch : int
            The batch size for generation. Increase for faster generation, at the cost of memory.
        gguf__n_cpu_threads : int
            The number of CPU threads to use for generation. Increase for much faster generation if multiple cores are available.
        save_history : bool
            Set to True if you want model to generate responses based on previous inputs (eg. chat history).
        system_prompt : str
            The system prompt to prepend to the input text.
        model_kwargs : Dict[str, Any]
            Additional keyword arguments for the model. These get passed directly to llama-cpp.Llama.__init__.
        """
        super().__init__(model_id, save_history, system_prompt)

        if Llama is None:
            raise ImportError(
                """llama-cpp not found. Install it using 'pip install llama-cpp-python'.
                              Please see https://github.com/abetlen/llama-cpp-python for more information."""
            )

        _model_kwargs = {
            "verbose": gguf__verbose,
            "n_ctx": gguf__n_ctx,
            "n_gpu_layers": gguf__n_gpu_layers,
            "n_batch": gguf__n_batch,
            "n_threads": gguf__n_cpu_threads,
            **(model_kwargs or {}),
        }

        check_invalid_input_parameters(Llama.__init__, _model_kwargs)

        if os.path.exists(model_id):
            self.llm = Llama(
                model_path=model_id,
                **_model_kwargs,
            )
        else:
            try:
                self.llm = Llama.from_pretrained(
                    repo_id=model_id,
                    filename=gguf__filename,
                    **_model_kwargs,
                )
            except ValueError as exc:
                error_msg = traceback.format_exc()
                if "Available Files:" in error_msg:
                    files_str = error_msg.split("Available Files:")[1].strip()
                    try:
                        files_list = literal_eval(files_str)
                        gguf_files = sorted(f for f in files_list if f.endswith(".gguf"))

                        logger.error("Specify a valid GGUF file in the 'gguf__filename' parameter")
                        logger.error("Available .gguf files:")
                        for file in gguf_files:
                            logger.error(file)
                    except (ValueError, SyntaxError):
                        logger.error("Could not parse available files list")
                raise InvalidGGUFFileError() from exc

    def generate(
        self,
        input_data: str,
        max_new_tokens: int = 512,
        temperature: float = 0.1,
        stopwords: Optional[List[str]] = None,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> str:
        templated_text, _generation_kwargs = self._prepare_generate(
            input_data, max_new_tokens, temperature, stopwords, generation_kwargs
        )

        generated_text = ""

        try:
            output = self.llm.create_chat_completion(templated_text, **_generation_kwargs)
            for item in output:
                new_text = item["choices"][0]["delta"].get("content", "")
                generated_text += new_text
                print(new_text, end="", flush=True)
        except KeyboardInterrupt:
            logger.warning("Control+C pressed, aborting generation")
        except Exception:
            logger.error(f"Error occured during generation: \n{traceback.format_exc()}")
        finally:
            print()  # Print newline after generation is done

        self.update_history(input_data, generated_text.strip())

        return generated_text.strip()

    def generate_stream(
        self,
        input_data: str,
        max_new_tokens: int = 512,
        temperature: float = 0.1,
        generation_kwargs: Optional[Dict[str, Any]] = None,
    ):
        templated_text, _generation_kwargs = self._prepare_generate(
            input_data, max_new_tokens, temperature, None, generation_kwargs
        )

        generated_text = ""
        try:
            output = self.llm.create_chat_completion(templated_text, **_generation_kwargs)
            for item in output:
                new_text = item["choices"][0]["delta"].get("content", "")
                generated_text += new_text
                yield new_text
        except KeyboardInterrupt:
            logger.warning("Control+C pressed, aborting generation")
        except Exception:
            logger.error(f"Error occured during generation: \n{traceback.format_exc()}")

        self.update_history(input_data, generated_text.strip())

    def _prepare_generate(self, input_data, max_new_tokens, temperature, stopwords, generation_kwargs):
        templated_text = self.apply_chat_template(input_data)

        _generation_kwargs = {
            "max_tokens": max_new_tokens,
            "temperature": temperature,
            "stream": True,
        }

        if stopwords:
            _generation_kwargs["stop"] = stopwords

        if generation_kwargs:
            _generation_kwargs.update(generation_kwargs)

        check_invalid_input_parameters(self.llm.create_chat_completion, _generation_kwargs)

        return templated_text, _generation_kwargs

    # override the original apply_chat_template method
    def apply_chat_template(self, input_data: str) -> List[Dict[str, str]]:
        messages = []
        if self.save_history:
            messages = self.history.copy()
        messages.append({"role": "user", "content": input_data})
        if self.system_prompt:
            messages.insert(0, {"role": "system", "content": self.system_prompt})

        return messages



from owlsight.app.run_app import run
from owlsight.processors.text_generation_manager import TextGenerationManager
from owlsight.ui.logo import print_logo
from owlsight.configurations.config_manager import ConfigManager
from owlsight.utils.deep_learning import check_gpu_and_cuda, calculate_max_parameters_per_dtype


def main():
    print_logo()
    check_gpu_and_cuda()
    calculate_max_parameters_per_dtype()

    config_manager = ConfigManager()
    text_generation_manager = TextGenerationManager(
        config_manager=config_manager,
    )

    # initialize agent
    run(text_generation_manager)


if __name__ == "__main__":
    main()



import sys
sys.path.append("src")

from owlsight.utils.helper_functions import extract_markdown


def test_extract_markdown():
    # Sample markdown input for Python, Bash, and CMD
    md_string = """
    ```python
    print("Hello, World!")
    ```

    ```bash
    echo "Hello, World!"
    ```

    ```cmd
    dir
    ```
    """

    # Expected output
    expected = [
        ("python", 'print("Hello, World!")'),
        ("bash", 'echo "Hello, World!"'),
        ("cmd", "dir"),
    ]

    result = extract_markdown(md_string)
    assert result == expected, f"Expected {expected}, but got {result}"



from typing import List


class ModuleNotFoundInVenvError(Exception):
    """Exception raised when a module is not found in the virtual environment."""

    def __init__(self, module_name: str, pyenv_path: str, lib_contents: List[str]):
        self.module_name = module_name
        self.pyenv_path = pyenv_path
        self.lib_contents = lib_contents
        self.message = f"Module '{module_name}' not found in '{pyenv_path}'. Contents of lib directory: {lib_contents}"
        super().__init__(self.message)


class QuantizationNotSupportedError(Exception):
    """Exception raised when quantization is not supported for a given model."""


class InvalidGGUFFileError(Exception):
    """Exception raised when the GGUF file is invalid."""
    


import sys
import os

sys.path.append("src")

import functools
import platform
import subprocess
import random
from typing import Optional
import logging
import traceback
import time
import shutil

import pytest
import psutil
from pynput.keyboard import Key, Controller

from owlsight.utils.constants import get_cache_dir

# Configure logging
SCRIPT = os.path.join("tests", "end_to_end_tests", "run_owlsight.py")
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

BUTTON_PRESS_TIME = 0.12

# Import pygetwindow only on Windows
if platform.system() == "Windows":
    try:
        import pygetwindow as gw
    except ImportError:
        subprocess.run(["pip", "install", "pygetwindow"], check=True)
        import pygetwindow as gw


def move_down_up(n):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.ensure_terminal_focused():
                raise RuntimeError("Could not focus terminal window")

            # Move down n times
            for _ in range(n):
                self.press_key(Key.down)

            # Execute the original function
            result = func(self, *args, **kwargs)

            # Move back up n times
            for _ in range(n):
                self.press_key(Key.up)

            return result

        return wrapper

    return decorator


class OwlsightStressTester:
    def __init__(self):
        self.keyboard = Controller()
        self.system = platform.system()
        self.owlsight_pid: Optional[int] = None
        self.max_startup_retries = 3
        self.window_wait_time = 2.0

        # Main menu options
        self.main_menu = ["how can I assist you?", "shell", "python", "config", "save", "load", "clear history", "quit"]
        self.in_main_menu = True
        self.main_menu_index = 0  # start at the top

        # Test commands for different modes
        self.python_commands = ["1+1", "print('test')", "owl_show()", "a=42"]
        self.shell_commands = ["pwd", "echo test", "ls", "dir"]
        self.ai_prompts = ["hi", "write a function", "help", "what is Python?"]

    def find_terminal_window(self) -> bool:
        """Find the terminal window with retries"""
        logger.debug("Looking for terminal window...")
        for attempt in range(3):
            if self.system == "Windows":
                windows = gw.getWindowsWithTitle("Owlsight-Terminal")
                if windows:
                    logger.debug(f"Found terminal window on attempt {attempt + 1}")
                    return True
            else:
                result = subprocess.run("xdotool search --name 'Owlsight-Terminal'", shell=True, capture_output=True)
                if result.returncode == 0:
                    logger.debug(f"Found terminal window on attempt {attempt + 1}")
                    return True

            logger.debug(f"Window not found, attempt {attempt + 1}, waiting...")
            time.sleep(1.0)

        return False

    def ensure_terminal_focused(self) -> bool:
        """Ensure the terminal window is focused before sending keyboard inputs"""
        for attempt in range(3):
            try:
                if self.system == "Windows":
                    windows = gw.getWindowsWithTitle("Owlsight-Terminal")
                    if not windows:
                        logger.error(f"Terminal window not found, attempt {attempt + 1}")
                        time.sleep(1.0)
                        continue

                    window = windows[0]
                    if not window.isActive:
                        window.activate()
                        time.sleep(0.5)  # Increased wait time after activation
                    return True
                else:
                    result = subprocess.run(
                        "xdotool search --name 'Owlsight-Terminal' windowactivate", shell=True, capture_output=True
                    )
                    if result.returncode == 0:
                        time.sleep(0.5)  # Increased wait time after activation
                        return True

                    logger.error(f"Failed to focus terminal window, attempt {attempt + 1}")
                    time.sleep(1.0)
            except Exception as e:
                logger.error(f"Error focusing terminal window: {e}")
                time.sleep(1.0)

        return False

    def set_main_menu_index(self, key: Key):
        """Set the main menu index based on key press"""
        if self.in_main_menu:
            if key == Key.down:
                self.main_menu_index = (self.main_menu_index + 1) % len(self.main_menu)
            elif key == Key.up:
                self.main_menu_index = (self.main_menu_index - 1) % len(self.main_menu)

    def get_main_menu_option(self) -> str:
        """Get the current main menu option"""
        return self.main_menu[self.main_menu_index]

    def find_owlsight_process(self) -> Optional[int]:
        """Find the Owlsight process"""
        for proc in psutil.process_iter(["pid", "name", "cmdline"]):
            try:
                if "owlsight" in str(proc.info["cmdline"]).lower() or "python" in str(proc.info["cmdline"]).lower():
                    return proc.info["pid"]
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
        return None

    def is_owlsight_running(self) -> bool:
        """Check if Owlsight process is still running"""
        if self.owlsight_pid:
            try:
                return psutil.pid_exists(self.owlsight_pid)
            except Exception as e:
                logger.error(f"Error checking if Owlsight is running: {e}")
        return False

    def type_fast(self, text: str):
        """Type text with minimal delay"""
        if not self.ensure_terminal_focused():
            raise RuntimeError("Could not focus terminal window")

        logger.debug(f"Typing text: {text}")
        for char in text:
            self.keyboard.press(char)
            self.keyboard.release(char)
        time.sleep(BUTTON_PRESS_TIME)

    def press_key(self, key: Key, times=1):
        """Press key with minimal delay"""
        if not self.ensure_terminal_focused():
            raise RuntimeError("Could not focus terminal window")

        logger.debug(f"Pressing key: {key} {times} times")
        for _ in range(times):
            self.keyboard.press(key)
            self.keyboard.release(key)
            self.set_main_menu_index(key)
        time.sleep(BUTTON_PRESS_TIME)

    @move_down_up(2)
    def test_python(self):
        """Test Python interpreter"""
        logger.debug("Starting test_python")
        if self.main_menu_index != 2:
            self._raise_wrong_mode_error("Python")
        self.press_key(Key.enter)
        time.sleep(BUTTON_PRESS_TIME)
        command = random.choice(self.python_commands)
        logger.debug(f"Executing Python command: {command}")
        self.type_fast(command)
        self.press_key(Key.enter)
        time.sleep(BUTTON_PRESS_TIME)
        self.type_fast("exit()")
        self.press_key(Key.enter)
        time.sleep(BUTTON_PRESS_TIME)

    @move_down_up(3)
    def test_rag(self):
        """Test RAG config"""
        logger.debug("Starting test_rag")
        if self.main_menu_index != 3:
            self._raise_wrong_mode_error("Config")
        self.press_key(Key.right, 3)
        self.press_key(Key.enter)
        # should be in rag menu now
        # activate rag
        self.press_key(Key.down)
        self.press_key(Key.right)
        self.press_key(Key.enter)
        self.press_key(Key.down, 2)
        self.type_fast("pandas")
        self.press_key(Key.enter)
        self.press_key(Key.down, 4)
        self.type_fast("How do I merge two dataframes?")
        # execute rag
        self.press_key(Key.enter)

        # deactivate rag
        self.press_key(Key.down)
        self.press_key(Key.right)
        self.press_key(Key.enter)

        # go back to main menu
        self.press_key(Key.enter)

    @move_down_up(1)
    def test_shell(self):
        """Test shell command execution"""
        logger.debug("Starting test_shell")
        if self.main_menu_index != 1:
            self._raise_wrong_mode_error("shell")
        self.press_key(Key.enter)
        time.sleep(BUTTON_PRESS_TIME)
        command = random.choice(self.shell_commands)
        logger.debug(f"Executing shell command: {command}")
        self.type_fast(command)
        self.press_key(Key.enter)
        time.sleep(BUTTON_PRESS_TIME)
        self.type_fast("exit")
        self.press_key(Key.enter)
        time.sleep(BUTTON_PRESS_TIME)

    def test_ai(self):
        """Test AI interaction"""
        logger.debug("Starting test_ai")
        if self.main_menu_index != 0:
            self._raise_wrong_mode_error("AI")
        prompt = random.choice(self.ai_prompts)
        logger.debug(f"Sending AI prompt: {prompt}")
        self.type_fast(prompt)
        self.press_key(Key.enter)
        time.sleep(0.2)

    def execute_random_action(self):
        """Execute a random action from available options"""
        if not self.ensure_terminal_focused():
            raise RuntimeError("Could not focus terminal window")

        actions = [
            (self.test_python, "python test"),
            (self.test_shell, "shell test"),
            (self.test_ai, "AI test"),
            (self.test_rag, "RAG test"),
        ]
        action, name = random.choice(actions)
        logger.debug(f"Selected action: {name}")

        try:
            action()
        except Exception as e:
            logger.error(f"Exception occurred during {name}: {traceback.format_exc()}")
            raise

        if not self.is_owlsight_running():
            logger.error(f"Owlsight process is not running after {name}")
            raise RuntimeError(f"Owlsight process died during {name}")

    def startup(self) -> bool:
        """Start Owlsight and verify it's running"""
        logger.debug("Starting up Owlsight")

        for attempt in range(self.max_startup_retries):
            logger.debug(f"Startup attempt {attempt + 1}/{self.max_startup_retries}")

            script_path = os.path.join(os.getcwd(), SCRIPT)

            # Start terminal
            if self.system == "Windows":
                cmd = f"start powershell -NoExit -Command \"cd '{os.getcwd()}'; $host.UI.RawUI.WindowTitle = 'Owlsight-Terminal'; python {script_path}\""
                subprocess.Popen(cmd, shell=True)
            else:
                subprocess.Popen(
                    [
                        "gnome-terminal",
                        "--working-directory",
                        os.getcwd(),
                        "--title=Owlsight-Terminal",
                        "--",
                        "python",
                        script_path,
                    ]
                )

            # Wait for terminal window to appear
            time.sleep(self.window_wait_time)

            if not self.find_terminal_window():
                logger.error(f"Terminal window not found on attempt {attempt + 1}")
                continue

            if not self.ensure_terminal_focused():
                logger.error(f"Failed to focus terminal window on attempt {attempt + 1}")
                continue

            logger.debug("Terminal window found and focused")

            time.sleep(5)  # Wait for Owlsight to start

            # Find process
            self.owlsight_pid = self.find_owlsight_process()
            if self.owlsight_pid:
                logger.debug(f"Owlsight started with PID: {self.owlsight_pid}")
                return True

            logger.error(f"Failed to find Owlsight process on attempt {attempt + 1}")

            # Cleanup before retry
            self.cleanup()
            time.sleep(1.0)

        logger.error("All startup attempts failed")
        return False

    def final_end_to_end_test(self) -> tuple[bool, str]:
        """Run final end-to-end verification"""
        logger.debug("Starting final_end_to_end_test")
        if not self.ensure_terminal_focused():
            raise RuntimeError("Could not focus terminal window")

        if self.main_menu_index != 0:
            self._raise_wrong_mode_error("AI")
        try:
            self.press_key(Key.up)
            time.sleep(0.05)

            # Execute quit
            self.press_key(Key.enter)
            time.sleep(1.0)

            return True, "Clean exit confirmed"

        except Exception as e:
            logger.error(f"End-to-end test failed: {e}")
            raise

    def cleanup(self):
        """Force cleanup if necessary"""
        logger.debug("Cleaning up")
        cache_dir = get_cache_dir()
        if os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)
        if self.is_owlsight_running():
            if self.system == "Windows":
                for window in gw.getWindowsWithTitle("Owlsight-Terminal"):
                    window.close()
                time.sleep(0.5)
                assert not gw.getWindowsWithTitle("Owlsight-Terminal"), "Failed to close terminal window"
            else:
                subprocess.run(["pkill", "-f", "Owlsight-Terminal"])
                time.sleep(0.5)
                assert not self.is_owlsight_running(), "Failed to kill Owlsight process"

    def _raise_wrong_mode_error(self, expected: str):
        actual = self.get_main_menu_option()
        logger.error(f"Wrong menu selected. Expected: {expected}, Actual: {actual}")
        raise RuntimeError(f"Wrong menu selected. Expected: {expected}, Actual: {actual}")


@pytest.mark.stress
def test_owlsight_stress():
    """
    Pytest function to run stress test on Owlsight.
    Tests stability through random operations and verifies clean exit.
    """
    num_iterations = 10  # Number of random operations to perform

    tester = OwlsightStressTester()

    try:
        # Start Owlsight
        startup_success = tester.startup()
        assert startup_success, "Failed to start Owlsight"

        # Run random tests
        for i in range(num_iterations):
            logger.debug(f"Starting iteration {i+1}/{num_iterations}")
            tester.execute_random_action()

        # Run end-to-end test
        success, message = tester.final_end_to_end_test()
        assert success, f"End-to-end test failed: {message}"

    finally:
        tester.cleanup()
        # assert not tester.is_owlsight_running(), "Owlsight process still running after cleanup"
        
if __name__ == "__main__":
    # Create run_owlsight.py if it doesn't exist
    run_script = """
import sys
sys.path.append("src")

from owlsight.main import main as m

if __name__ == "__main__":
    m()
    """

    with open(SCRIPT, "w") as f:
        f.write(run_script)

    # Allow running directly for debugging
    pytest.main([__file__, "-v", "-s"])



import time
import sys

sys.path.append("src")

import pytest
from transformers import AutoTokenizer
from owlsight.processors.text_generation_processors import (
    TextGenerationProcessorTransformers,
)


@pytest.fixture
def setup_processor():
    """Fixture to set up the text generation processor and tokenizer."""
    model_id = "hf-internal-testing/tiny-random-GPTNeoXForCausalLM"
    tokenizer = AutoTokenizer.from_pretrained(model_id)
    processor = TextGenerationProcessorTransformers(model_id)
    return processor, tokenizer


def test_generate_response(setup_processor):
    """Test that the processor generates a valid response."""
    processor, _ = setup_processor
    prompt = "test prompt"
    max_new_tokens = 128

    response = processor.generate(prompt, max_new_tokens=max_new_tokens)
    assert isinstance(response, str), "Generated response should be a string."


def test_token_count_within_tolerance(setup_processor):
    """Test that the generated token count is within the acceptable range."""
    processor, tokenizer = setup_processor
    prompt = "test prompt"
    max_new_tokens = 128
    tolerance_fraction = 0.5

    response = processor.generate(prompt, max_new_tokens=max_new_tokens)

    # Tokenize response and calculate new tokens
    response_tokens = tokenizer.tokenize(response)
    prompt_tokens = tokenizer.tokenize(prompt)
    new_tokens = response_tokens[len(prompt_tokens) :]

    # Calculate acceptable range
    lower_bound = max_new_tokens - (max_new_tokens * tolerance_fraction)
    upper_bound = max_new_tokens + (max_new_tokens * tolerance_fraction)

    assert lower_bound <= len(new_tokens) <= upper_bound, (
        f"Expected approximately {max_new_tokens} tokens "
        f"(range: {lower_bound:.2f} - {upper_bound:.2f}), got {len(new_tokens)}"
    )


def test_prompt_tokens_exclusion(setup_processor):
    """Test that the response excludes prompt tokens when counting new tokens."""
    processor, tokenizer = setup_processor
    prompt = "test prompt"
    max_new_tokens = 128

    response = processor.generate(prompt, max_new_tokens=max_new_tokens)

    # Tokenize response and calculate new tokens
    response_tokens = tokenizer.tokenize(response)
    prompt_tokens = tokenizer.tokenize(prompt)

    assert len(response_tokens) > len(prompt_tokens), "Response tokens should exceed prompt tokens."


def test_invalid_generation_kwargs(setup_processor):
    """Test that the processor handles invalid kwargs without freezing."""
    processor, _ = setup_processor
    prompt = "test prompt"
    max_new_tokens = 128
    generation_kwargs = {"non_existing_kwarg": "value"}

    # Record start time
    start_time = time.time()

    try:
        # Call generate method and ensure it returns within a reasonable time
        response = processor.generate(prompt, max_new_tokens=max_new_tokens, generation_kwargs=generation_kwargs)
    except Exception as exc:
        # If an exception is raised, ensure it's not due to a freeze
        duration = time.time() - start_time
        assert duration < 5, "The generate method froze."
        assert "non_existing_kwarg" in str(exc), "Error message should contain the invalid keyword."
        return  # Test passes as the exception was correctly raised

    # If no exception is raised, ensure the response is as expected
    duration = time.time() - start_time
    assert duration < 5, "The generate method froze."
    assert response == "", "Response should be empty for invalid generation kwargs."


def test_num_beams_with_streaming_raises_error(setup_processor):
    """Test that using num_beams with streaming produces appropriate error output."""
    processor, _ = setup_processor
    prompt = "test prompt"
    max_new_tokens = 128
    num_beams = 4

    # Generate with num_beams (will print error in thread)
    with pytest.raises(ValueError) as exc:
        processor.generate(prompt, max_new_tokens=max_new_tokens, generation_kwargs={"num_beams": num_beams})
        assert "num_beams" in str(exc.value), "Error message should mention num_beams parameter"


if __name__ == "__main__":
    pytest.main([__file__])



import tempfile
import traceback
from typing import Union
from enum import Enum, auto
import os

from owlsight.ui.file_dialogs import save_file_dialog, open_file_dialog
from owlsight.ui.console import get_user_choice, print_colored, get_user_input
from owlsight.processors.text_generation_manager import TextGenerationManager
from owlsight.app.handlers import handle_interactive_code_execution
from owlsight.utils.code_execution import CodeExecutor, execute_code_with_feedback
from owlsight.utils.helper_functions import (
    force_delete,
    remove_temp_directories,
    parse_media_placeholders,
    os_is_windows,
)
from owlsight.utils.venv_manager import get_lib_path, get_pip_path, get_pyenv_path, get_temp_dir
from owlsight.utils.constants import (
    PROMPT_COLOR,
    MAIN_MENU,
    get_cache_dir,
    get_pickle_cache,
    get_prompt_cache,
    get_py_cache,
)
from owlsight.utils.deep_learning import free_memory
from owlsight.rag.python_lib_search import search_python_libs
from owlsight.utils.logger import logger


class CommandResult(Enum):
    """Enum to represent the result of a command from the mainmenu."""
    CONTINUE = auto()
    BREAK = auto()
    PROCEED = auto()


def run_code_generation_loop(code_executor: CodeExecutor, manager: TextGenerationManager) -> None:
    """Runs the main loop for code generation and user interaction."""
    option = None
    user_choice = None
    while True:
        try:
            print_colored("Make a choice:", color=PROMPT_COLOR)

            # define startindex of arrow in mainmenu
            _option_or_userchoice: bool = option or user_choice
            if _option_or_userchoice:
                start_index = list(MAIN_MENU.keys()).index(_option_or_userchoice)
            else:
                start_index = 0
            user_choice, option = get_user_input(start_index=start_index)

            if not user_choice and option not in ["config", "save", "load"]:
                logger.error("User choice is empty. Please try again.")
                continue

            command_result = handle_special_commands(option, user_choice, code_executor, manager)
            if command_result == CommandResult.BREAK:
                break
            elif command_result == CommandResult.CONTINUE:
                continue

            if manager.processor is None:
                logger.error("Processor not set. Please load a model first by setting 'model.model_id' in the config!")
                continue
            else:
                process_user_question(user_choice, code_executor, manager)

        except KeyboardInterrupt:
            logger.info("KeyboardInterrupt received. Restarting...")
        except Exception:
            logger.error(f"Unexpected error:\n{traceback.format_exc()}")
            # raise


def handle_special_commands(
    choice_key: Union[str, None],
    user_choice: str,
    code_executor: CodeExecutor,
    manager: TextGenerationManager,
) -> CommandResult:
    """Handles special commands such as shell, config, save, load, python, clear history, and quit."""
    if choice_key == "shell":
        code_executor.execute_code_block(lang=choice_key, code_block=user_choice)
        return CommandResult.CONTINUE
    elif choice_key == "config":
        config_key = ""
        while not config_key.endswith("back"):
            config_key = handle_config_update(user_choice, manager)
        return CommandResult.CONTINUE
    elif choice_key == "save":
        if not user_choice and os_is_windows():
            file_path = save_file_dialog(initial_dir=os.getcwd(), default_filename="owlsight_config.json")
            if not file_path:
                logger.error("No file selected. Please try again.")
                return CommandResult.CONTINUE
            user_choice = file_path
        manager.save_config(user_choice)
        return CommandResult.CONTINUE
    elif choice_key == "load":
        if not user_choice and os_is_windows():
            file_path = open_file_dialog(initial_dir=os.getcwd())
            if not file_path:
                logger.error("No file selected. Please try again.")
                return CommandResult.CONTINUE
            user_choice = file_path
        manager.load_config(user_choice)
        return CommandResult.CONTINUE
    elif user_choice == "python":
        handle_interactive_code_execution(code_executor)
        return CommandResult.CONTINUE
    elif user_choice == "clear history":
        clear_history(code_executor, manager)
        return CommandResult.CONTINUE
    elif user_choice == "quit":
        logger.info("Quitting...")
        return CommandResult.BREAK
    return CommandResult.PROCEED


def handle_config_update(user_choice: str, manager: TextGenerationManager) -> str:
    """Handles updating the configuration based on the user's choice."""
    logger.info(f"Chosen config: {user_choice}")

    # Retrieve nested configuration options
    available_choices = manager.get_config_choices()
    selected_config = available_choices[user_choice]

    # Get user choice for the nested configuration
    user_selected_choice = get_user_choice(selected_config, return_value_only=False)

    if isinstance(user_selected_choice, dict):
        nested_key = next(iter(user_selected_choice))  # Get the first key
        config_value = user_selected_choice[nested_key]  # Get the corresponding value
    else:
        nested_key = user_selected_choice
        config_value = None

    # Construct the config key and update the configuration
    config_key = f"{user_choice}.{nested_key}"
    manager.update_config(config_key, config_value)

    return config_key


def clear_history(code_executor: CodeExecutor, manager: TextGenerationManager) -> None:
    """Clears the following things:

    - All variables in the Python interpreter state, except those starting with "owl_"
    - Python interpreter history file
    - Prompt history file
    - chat history in the processor
    - pickled cache files
    """
    # clear all variables except those starting with "owl_"
    temp_dict = {k: v for k, v in code_executor.globals_dict.items() if k.startswith("owl_")}
    code_executor.globals_dict.clear()
    code_executor.globals_dict.update(temp_dict)

    force_delete(get_cache_dir())

    if manager.processor is not None:
        manager.processor.history.clear()

    logger.info(f"Cleared cachefolder {get_cache_dir()} and model chathistory.")

    # rebuild empty cache files after clearing
    get_pickle_cache()
    get_prompt_cache()
    get_py_cache()


def process_user_question(user_choice: str, code_executor: CodeExecutor, manager: TextGenerationManager) -> None:
    # Parse media placeholders in the user choice, if present. 
    user_question, media_objects = parse_media_placeholders(user_choice, code_executor.globals_dict)
    rag_is_active = manager.get_config_key("rag.active", False)
    library_to_rag = manager.get_config_key("rag.target_library", "")
    if rag_is_active and library_to_rag:
        logger.info(f"RAG search enabled. Adding context of python library '{library_to_rag}' to the question.")
        ctx_to_add = f"""
# CONTEXT:
The following context is documentation from the python library {library_to_rag}.
Use this information to help generate a code snippet that answers the question.
"""
        context = search_python_libs(
            library_to_rag, user_question, manager.get_config_key("top_k", 3), cache_dir=get_pickle_cache()
        )
        ctx_to_add += context
        user_question = f"{user_question}\n\n{ctx_to_add}".strip()
        logger.info(f"Context added to the question with approx amount of {len(context.split())} words")

    response = manager.generate(user_question, media_objects=media_objects)
    execute_code_with_feedback(
        response=response,
        original_question=user_question,
        code_executor=code_executor,
        prompt_code_execution=manager.config_manager.get("main.prompt_code_execution", True),
        prompt_retry_on_error=manager.config_manager.get("main.prompt_retry_on_error", False),
    )


def run(manager: TextGenerationManager) -> None:
    """
    Main function to run the interactive loop for code generation and execution

    Parameters
    ----------
    manager : TextGenerationManager
        TextGenerationManager instance to handle the code generation and execution
    """
    pyenv_path = get_pyenv_path()
    lib_path = get_lib_path(pyenv_path)
    pip_path = get_pip_path(pyenv_path)

    # Remove lingering temporary directories
    remove_temp_directories(lib_path)

    temp_dir_location = get_temp_dir(".owlsight_packages")

    # Create temporary directory in venv to install packages, until end of execution lifecycle
    with tempfile.TemporaryDirectory(dir=temp_dir_location) as temp_dir:
        logger.info(f"Temporary directory created at: {temp_dir}")

        code_executor = CodeExecutor(manager, pyenv_path, pip_path, temp_dir)

        run_code_generation_loop(code_executor, manager)

    logger.info(f"Removing temporary directory: {temp_dir}")
    free_memory()
    force_delete(temp_dir)



"""This module contains a function to check if Tesseract is installed on the system."""

import ctypes
import os
import platform
import subprocess
import traceback
from urllib.parse import urlparse

import pytesseract
import requests
from bs4 import BeautifulSoup
from tqdm import tqdm

from owlsight.utils.helper_functions import os_is_windows
from owlsight.utils.logger import logger


def setup_tesseract() -> str:
    """Initialize Tesseract. Return the path to the Tesseract executable."""
    if not os_is_windows():
        logger.error("Automatic installation is currently only supported on Windows.")
        return ""
    TESSERACT_LOCATION = find_tesseract_installation()
    if TESSERACT_LOCATION:
        logger.info("Tesseract is installed at %s.", TESSERACT_LOCATION)
        pytesseract.pytesseract.tesseract_cmd = TESSERACT_LOCATION
    else:
        logger.info("Tesseract not found. Downloading and installing...")
        download_and_install_tesseract()
        TESSERACT_LOCATION = find_tesseract_installation()
        if TESSERACT_LOCATION:
            logger.info("Tesseract is now installed at %s.", TESSERACT_LOCATION)
            pytesseract.pytesseract.tesseract_cmd = TESSERACT_LOCATION
        else:
            url = "https://github.com/UB-Mannheim/tesseract/wiki"
            logger.error("Failed to install Tesseract! You may download it manually from %s", url)
            raise RuntimeError(f"Failed to install Tesseract! You may download it manually from {url}")

    return TESSERACT_LOCATION


def download_and_install_tesseract():
    """Download and install Tesseract if not found."""
    output_dir = "."
    downloaded_file = download_tesseract(output_dir)
    if downloaded_file:
        exe_path = os.path.join(output_dir, downloaded_file)
        logger.info("Installing Tesseract from %s...", exe_path)
        install_tesseract(exe_path)
        os.remove(exe_path)
        logger.info("Removed the Tesseract installer %s", exe_path)


def find_tesseract_installation() -> str:
    """Find and return the path to the Tesseract executable."""
    tesseract_path = ""
    if os_is_windows():
        common_paths = [
            r"C:\Program Files\Tesseract-OCR\tesseract.exe",
            r"C:\Program Files (x86)\Tesseract-OCR\tesseract.exe",
        ]
    else:
        common_paths = ["/usr/bin/tesseract", "/usr/local/bin/tesseract"]

    for path in common_paths:
        if os.path.exists(path):
            return path
    try:
        subprocess.run(
            ["tesseract", "--version"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return "tesseract"
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass
    return tesseract_path


def download_tesseract(output_dir=".", timeout=10) -> str | None:
    """
    Download Tesseract .exe from the specified GitHub page and show a progress bar during download.

    Parameters
    ----------
    output_dir : str
        The directory to save the downloaded file to.

    Returns
    -------
    str | None
        The name of the downloaded file, or None if the download failed.
    """
    tesseract_url = "https://github.com/UB-Mannheim/tesseract/wiki"
    try:
        res = requests.get(tesseract_url, timeout=timeout)
        res.raise_for_status()  # Raises an HTTPError if the response status code is 4XX/5XX
        soup = BeautifulSoup(res.text, "html.parser")
        for link in soup.find_all("a"):
            href = link.get("href", "")
            if "tesseract-ocr-w64-setup" in href and href.endswith(".exe"):
                logger.info("Found an exe file online called %s", href)
                file_name = os.path.basename(urlparse(href).path)
                file_path = os.path.join(output_dir, file_name)
                if not os.path.exists(file_path):
                    try:
                        logger.info("Downloading %s...", file_name)
                        with requests.get(href, stream=True, timeout=timeout) as exe_response:
                            exe_response.raise_for_status()
                            total_size_in_bytes = int(exe_response.headers.get("content-length", 0))
                            block_size = 1024  # 1 Kibibyte
                            progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
                            with open(file_path, "wb") as file:
                                for data in exe_response.iter_content(block_size):
                                    progress_bar.update(len(data))
                                    file.write(data)
                            progress_bar.close()
                        logger.info("Successfully downloaded %s", file_name)
                        return file_name
                    except Exception as e:
                        logger.error("Download interrupted, removing partially downloaded file.")
                        os.remove(file_path)  # Remove the partially downloaded file
                        raise e
                else:
                    logger.info("File %s already exists.", file_path)
                    return os.path.basename(file_path)
        logger.error("Failed to find the download link.")
    except requests.HTTPError as http_err:
        logger.error(f"HTTP error occurred: {http_err}")
    except Exception as err:
        logger.error(f"An error occurred: {err}")
    return None


def is_admin():
    """Check if the script is running with administrative privileges."""
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except:
        return False


def install_tesseract(exe_path: str):
    """Install Tesseract by running the downloaded .exe installer."""
    if platform.system() == "Windows":
        if is_admin():
            try:
                subprocess.run([exe_path, "/S"], check=True)
            except subprocess.CalledProcessError:
                logger.error(
                    "CalledProcessError while installing Tesseract:\n%s",
                    traceback.format_exc(),
                )
            except Exception:
                logger.error(
                    "An unexpected error occurred during Tesseract installation:\n%s",
                    traceback.format_exc(),
                )
        else:
            raise OSError("The script is not running with administrative privileges. Please run as administrator.")
    else:
        logger.error("Automatic installation is currently only supported on Windows.")



import os
import pytest
import shutil
from unittest.mock import patch
import tempfile
import sys

sys.path.append("src")
from owlsight.utils.constants import create_directory, create_file
from owlsight.utils.helper_functions import os_is_windows


@pytest.fixture
def test_env():
    """Create and manage test environment using tempfile for cross-platform compatibility."""
    # Create temporary directory using tempfile (handles cross-platform paths)
    temp_dir = tempfile.mkdtemp()
    test_home = os.path.join(temp_dir, "home")
    test_cache_dir = os.path.join(test_home, ".owlsight")

    # Create temporary home directory
    os.makedirs(test_home, exist_ok=True)
    os.makedirs(test_cache_dir, exist_ok=True)

    # Create a fixture return value with all needed test data
    env_data = {
        "test_home": test_home,
        "test_cache_dir": test_cache_dir,
        "test_py_cache": os.path.join(test_cache_dir, ".python_history"),
        "test_prompt_cache": os.path.join(test_cache_dir, ".prompt_history"),
    }

    # Use context manager for patching
    with patch("os.path.expanduser", return_value=test_home):
        yield env_data

    # Cleanup after tests
    shutil.rmtree(temp_dir)


def test_create_directory(test_env):
    """Test that create_or_get_path creates subdirectories."""
    test_path = "test_subdir"
    cache_dir = test_env["test_cache_dir"]
    full_test_path = os.path.join(cache_dir, test_path)
    result = str(create_directory(test_path, base=cache_dir))

    # Use normpath to handle path separators consistently across platforms
    actual_full_path = os.path.normpath(os.path.join(cache_dir, full_test_path))

    assert result == full_test_path
    assert os.path.exists(actual_full_path), f"Path does not exist: {actual_full_path}"
    assert os.path.isdir(actual_full_path)


def test_create_file(test_env):
    """Test that test_create_file creates files."""
    test_path = "test_file"
    cache_dir = test_env["test_cache_dir"]
    full_test_path = os.path.join(cache_dir, test_path)
    result = str(create_file(test_path, base=cache_dir))

    # Use normpath to handle path separators consistently across platforms
    actual_full_path = os.path.normpath(os.path.join(cache_dir, full_test_path))

    assert result == full_test_path
    assert os.path.exists(actual_full_path), f"File does not exist: {actual_full_path}"
    assert os.path.isfile(actual_full_path)


def test_create_directory_returns_full_path(test_env):
    """Test that create_directory returns the full absolute path"""
    test_path = ".test_history"
    cache_dir = test_env["test_cache_dir"]
    result = str(create_directory(test_path, base=cache_dir))

    # Result should be a full path
    assert os.path.isabs(result), "Should return absolute path"
    assert result.endswith(test_path), "Should contain the input path"
    assert result.startswith(cache_dir), "Should start with cache dir"

    # The directory should exist at the full path
    assert os.path.exists(result)
    assert os.path.isdir(result)


def test_get_cache_dir_permissions(test_env):
    """Test that created directories have correct permissions."""
    cache_dir = test_env["test_cache_dir"]
    mode = os.stat(cache_dir).st_mode & 0o777

    if os_is_windows():
        # Windows permissions are different, typically 666 or 777
        assert mode & 0o400  # Check if readable
        assert mode & 0o200  # Check if writable
    else:
        # Unix permissions
        assert mode in (0o755, 0o775)


def test_path_separators(test_env):
    """Test that path separators are handled correctly for the current platform."""
    cache_dir = test_env["test_cache_dir"]
    test_path = os.path.join("nested", "path")
    full_test_path = os.path.join(cache_dir, test_path)
    result = str(create_directory(full_test_path, base=cache_dir))

    # Adjust expected path to match OS behavior
    expected_path = os.path.normpath(full_test_path)
    assert os.path.normpath(result) == expected_path, f"Unexpected path formatting: {result}"

if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



"""
Created by Nestor Demeure.
This software is released under the Apache License 2.0.
"""

import torch
from transformers import StoppingCriteria, AutoTokenizer
from typing import List, Literal, Dict
from dataclasses import dataclass, field


class StopWordCriteria(StoppingCriteria):
    """
    A stopping criteria that halts the text generation process if any specified stop word is encountered.

    Inspired by https://discuss.huggingface.co/t/implimentation-of-stopping-criteria-list/20040/9
    And: https://github.com/outlines-dev/outlines/blob/main/outlines/generate/api.py
    """

    def __init__(
        self,
        tokenizer: AutoTokenizer,
        prompts: List[str],
        stop_words: List[str] = [],
        check_every: int = 1,
    ):
        """
        Initializes the StopWordCriteria with the necessary parameters for checking stop words during text generation.

        Parameters:
        ----------
            tokenizer (AutoTokenizer): The tokenizer for encoding prompts and stop words.
            prompts (List[str]): Initial prompts used for generation, needed to determine where generated text begins.
            stop_words (List[str]): Words that trigger the stopping of generation when detected.
            check_every (int): Frequency of checking for stop words in the token stream (a performance optimization, use 1 to cut it out).
        """
        super().__init__()
        self.tokenizer = tokenizer
        self.input_sizes = [self.tokenizer.encode(prompt, return_tensors="pt").size(-1) for prompt in prompts]
        self.stop_words = stop_words
        self.max_stop_word_size = max(
            (self.tokenizer.encode(word, return_tensors="pt").size(-1) for word in stop_words),
            default=0,
        )
        self.check_every = check_every

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs) -> bool:
        """
        Determines whether to stop generation based on the presence of stop words.

        Stops if a stop word is found in *all* batch elements *and* the sequence length is a multiple of `check_every`.
        Note: Delay in stopping may occur if `check_every > 1`.

        Parameters:
            input_ids (torch.LongTensor): Generated token IDs.
            scores (torch.FloatTensor): Generation scores for each token. Not used here.

        Returns:
            bool: True to stop generation, False to continue.
        """
        batch_size, seq_len = input_ids.shape

        # Skip check if no stop words are defined or it is not yet time to check
        if (len(self.stop_words) == 0) or (seq_len % self.check_every != 0):
            return False

        for i in range(batch_size):
            # Calculate starting index for new tokens
            prompt_size = self.input_sizes[i]
            max_new_tokens = (2 * self.max_stop_word_size) + self.check_every
            latest_tokens = input_ids[i, prompt_size:][-max_new_tokens:]

            # Check for stop words in the decoded text
            if not any(
                word in self.tokenizer.decode(latest_tokens, skip_special_tokens=True) for word in self.stop_words
            ):
                return False  # Continue generation if any batch item lacks stop words

        return True  # Stop generation if all conditions are met

    def extract_answers(self, input_ids: torch.LongTensor, strip_stopword: bool = True) -> List[str]:
        """
        Extracts generated answers by removing prompts and optionally stopping at the first stop word.

        Parameters:
            input_ids (torch.LongTensor): Generated token IDs.
            strip_stopword (bool): Determines whether the stop word is removed from the output.

        Returns:
            List[str]: Extracted answers, with or without stop words.
        """
        batch_size, _ = input_ids.shape
        result = []

        for i in range(batch_size):
            # Decode generated tokens to text, excluding the prompt
            prompt_size = self.input_sizes[i]
            answer_tokens = input_ids[i, prompt_size:]
            answer_text = self.tokenizer.decode(answer_tokens, skip_special_tokens=True)

            # Find the first occurrence of any stop word
            lower_stop_index = len(answer_text)  # Default to end of text
            for word in self.stop_words:
                stop_index = answer_text.find(word)
                if stop_index != -1:
                    # Adjust stop index based on whether we're stripping the stop word
                    stop_index += 0 if strip_stopword else len(word)
                    lower_stop_index = min(stop_index, lower_stop_index)

            # Cut the text at the first stop word found (if any)
            answer_text = answer_text[:lower_stop_index]
            result.append(answer_text)

        return result

    def __len__(self):
        return 1

    def __iter__(self):
        yield self


class SingletonDict(dict):
    """
    A dictionary that only allows a single instance to be created.
    Meant to be used as a singleton for storing global variables and to share state across different classes.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(SingletonDict, cls).__new__(cls, *args, **kwargs)
        return cls._instance


MediaType = Literal["image", "audio", "video"]


@dataclass
class MediaObject:
    """
    Represents a media object with its type, path, and options.

    Attributes
    ----------
    type : MediaType
        The type of media (image, audio, or video)
    path : str
        The path to the media file
    options : Dict[str, str]
        Optional parameters for processing the media
    """

    type: MediaType
    path: str
    options: Dict[str, str] = field(default_factory=lambda: {}) # ensure each instance has its own options dict



from typing import List, Tuple, Dict, Any
import ast
import os
import shutil
import re
import traceback
import inspect
from datetime import datetime, timedelta

from prompt_toolkit import prompt
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.styles import Style

from owlsight.utils.logger import logger
from owlsight.utils.custom_classes import MediaObject, MediaType


def extract_markdown(md_string: str) -> List[Tuple[str, str]]:
    """
    Extract language and code blocks from a markdown string.
    """
    pattern = r"```(\w+)([\s\S]*?)```"
    return [(match[0].strip(), match[1].strip()) for match in re.findall(pattern, md_string)]


def replace_bracket_placeholders(text: str, var_dict: Dict[str, Any]) -> Any:
    """
    Evaluates expressions inside {{...}} in the given text and replaces them with the result.
    Correctly handles expressions containing braces or other special characters.

    Parameters
    ----------
    text : str
        The input string containing placeholders in the form of `{{...}}`.
    var_dict : dict
        A dictionary where keys correspond to variables used in placeholders.

    Returns
    -------
    Any
        The evaluated object if the entire string is a single placeholder,
        otherwise the string with placeholders replaced.
    """

    def evaluate_expression(expr: str) -> Any:
        try:
            safe_globals = {
                "__builtins__": {
                    "abs": abs,
                    "all": all,
                    "any": any,
                    "bin": bin,
                    "bool": bool,
                    "chr": chr,
                    "divmod": divmod,
                    "enumerate": enumerate,
                    "filter": filter,
                    "float": float,
                    "format": format,
                    "hash": hash,
                    "hex": hex,
                    "int": int,
                    "isinstance": isinstance,
                    "issubclass": issubclass,
                    "iter": iter,
                    "len": len,
                    "list": list,
                    "map": map,
                    "max": max,
                    "min": min,
                    "next": next,
                    "oct": oct,
                    "ord": ord,
                    "pow": pow,
                    "range": range,
                    "repr": repr,
                    "reversed": reversed,
                    "round": round,
                    "sorted": sorted,
                    "str": str,
                    "sum": sum,
                    "tuple": tuple,
                    "zip": zip,
                    "dict": dict,
                    "set": set,
                    "frozenset": frozenset,
                    "datetime": datetime,
                    "timedelta": timedelta,
                }
            }
            safe_globals.update(var_dict)
            return eval(expr, safe_globals, {})
        except Exception as e:
            error_message = f"Error evaluating '{expr}': {str(e)}"
            raise type(e)(error_message) from None

    # Pattern to match balanced double braces
    pattern = r"""
        \{\{            # Opening double braces {{
        (?P<expr>       # Start of named group 'expr'
            [^\{\}]*    # Match any characters except { or }
            (?:         # Non-capturing group
                \{[^\{\}]*\} # Match balanced braces
                [^\{\}]*    # Match any characters except { or }
            )*          # Zero or more times
        )               # End of named group 'expr'
        \}\}            # Closing double braces }}
    """
    regex = re.compile(pattern, re.VERBOSE)

    # Function to replace each placeholder
    def replace_match(m):
        expr = m.group("expr")
        evaluated = evaluate_expression(expr.strip())
        return str(evaluated)

    # Check if the entire text is a single placeholder
    if regex.fullmatch(text):
        expr = regex.fullmatch(text).group("expr")
        return evaluate_expression(expr.strip())
    else:
        return regex.sub(replace_match, text)


def editable_input(prompt_text: str, default_value: str, color: str = "ansicyan") -> str:
    """
    Displays a prompt with a pre-filled editable string and custom color for the default value.

    Parameters
    ----------
    prompt_text : str
        The prompt message shown before the editable string.
    default_value : str
        The string that will be pre-filled and editable by the user.
    color : str, optional
        The color to apply to the default value in the prompt message, default is 'ansicyan'.

    Examples
    --------
    >>> editable_input("Enter your name: ", "John")
    Enter your name: "John" -> Enter your name: "Johnny"
    'Johnny'

    Returns
    -------
    str
        The string edited by the user.
    """
    style = Style.from_dict({"prompt_text": color})

    # Prepare the prompt text with custom color using HTML
    formatted_prompt = HTML(f"<ansicyan>{prompt_text}</ansicyan>")

    # Get the result from the prompt (default value is shown but not styled)
    result = prompt(formatted_prompt, default=default_value, style=style)

    return result.strip()


def force_delete(temp_dir: str) -> None:
    """Forcefully deletes a directory if it exists."""
    if os.path.exists(temp_dir):
        try:
            shutil.rmtree(temp_dir)
        except Exception:
            logger.error(f"Error deleting directory {temp_dir}:\n{traceback.format_exc()}")


def remove_temp_directories(lib_path: str) -> None:
    """Removes lingering temporary directories in the virtual environment's library path."""
    for d in os.listdir(lib_path):
        if d.startswith("tmp"):
            logger.info(f"Removing temporary directory: {d}")
            force_delete(os.path.join(lib_path, d))


def format_error_message(e: Exception) -> str:
    """
    Format an error message to be displayed to the user.

    Parameters
    ----------
    error : Exception
        The exception that occurred.

    Returns
    -------
    str
        The formatted error message.
    """
    return "{e.__class__.__name__}: {e}".format(e=e)


def convert_to_real_type(value):
    """
    Convert a string to its real type if possible (e.g., 'True' -> True, '3.14' -> 3.14).
    """
    if not isinstance(value, str):
        return value

    # Try to evaluate the string and return the result only if it's not a string
    try:
        evaluated_value = ast.literal_eval(value)
        # Only return the evaluated value if it is not a string
        if not isinstance(evaluated_value, str):
            return evaluated_value
    except (ValueError, SyntaxError):
        pass  # Return original string if evaluation fails

    return value  # Return the original string if it's not evaluable


def os_is_windows():
    return os.name == "nt"


def check_invalid_input_parameters(func: callable, kwargs: dict):
    """
    Validate the keyword arguments passed to a class against the __init__ signature.

    Parameters
    ----------
    func : callable
        The callable of which arguments are being validated.
    kwargs : dict
        A dictionary of keyword arguments to validate.

    Raises
    ------
    ValueError
        If there are invalid parameters.
    """
    # Extract the parameters from the __init__ method of the class
    sig = inspect.signature(func)
    sig_params = sig.parameters

    valid_params = [param_name for param_name in sig_params if param_name != "self"]

    # Check for any extra parameters in kwargs that are not in the __init__ signature
    for key in kwargs:
        if key not in sig_params:
            raise ValueError(
                f"Invalid argument: '{key}' is not a valid parameter for '{func.__name__}'\nValid parameters: {valid_params}"
            )


def flatten_dict(d, parent_key="", sep=".") -> dict:
    """Flatten a nested dictionary."""
    flattened = {}
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            flattened.update(flatten_dict(v, new_key, sep=sep))
        else:
            flattened[new_key] = v
    return flattened


def parse_media_placeholders(text: str, var_dict: Dict[str, Any]) -> Tuple[str, Dict[str, MediaObject]]:
    """
    Parse media syntax patterns [[type:path|option1=value1|...]] in text and evaluate any
    Python expressions inside {{...}}. Returns the modified text and a dictionary of
    media objects with their options.

    Parameters
    ----------
    text : str
        The input string containing media placeholders and optional Python expressions
    var_dict : dict
        A dictionary where keys correspond to variables used in Python expressions

    Returns
    -------
    Tuple[str, MediaObjects]
        A tuple containing:
        - The text with media placeholders replaced with unique identifiers
        - A dictionary mapping identifiers to MediaObject instances

    Examples
    --------
    >>> var_dict = {"folder": "images", "filename": "cat.jpg"}
    >>> text = "Analyze this: [[image:{{folder}}/{{filename}}|width=512]]"
    >>> result, media_objects = parse_media_placeholders(text, var_dict)
    >>> print(result)
    'Analyze this: __MEDIA_0__'
    >>> print(media_objects)
    {
        '__MEDIA_0__': MediaObject(
            type='image',
            path='images/cat.jpg',
            options={'width': '512'}
        )
    }
    """

    def validate_media_syntax(text: str) -> None:
        # Check for valid media types
        invalid_types = re.findall(r"\[\[(\w+):", text)
        valid_types = {"image", "audio", "video"}
        for t in invalid_types:
            if t not in valid_types:
                raise ValueError(f"Invalid media type: {t}. Must be one of {valid_types}")

        # Check for missing paths
        if re.search(r"\[\[\w+:\s*(\||\]\])", text):
            raise ValueError("Media path cannot be empty")

        # Check for invalid option format
        option_pattern = r"\|(?!\w+=)[^]|]*(?=[\]|])"
        invalid_options = re.findall(option_pattern, text)
        if invalid_options:
            raise ValueError(f"Invalid option format: {invalid_options[0].strip()}. Must be key=value")

    validate_media_syntax(text)

    pattern = r"""\[\[
        (?P<type>image|audio|video):  # Media type
        (?P<path>[^\|\]]+)            # Path (anything until | or ])
        (?:\|(?P<options>[^\]]+))?    # Optional options after |
        \]\]"""

    media_objects: Dict[str, MediaObject] = {}
    replacement_count = 0

    def replace_match(match) -> str:
        nonlocal replacement_count

        media_type: MediaType = match.group("type")  # type: ignore
        raw_path = match.group("path")
        options_str = match.group("options") or ""

        # Process the path first - evaluate any Python expressions
        processed_path = replace_bracket_placeholders(raw_path, var_dict)

        # Process options
        options: Dict[str, str] = {}
        if options_str:
            for option in options_str.split("|"):
                if "=" in option:
                    key, value = option.split("=", 1)
                    # Evaluate Python expressions in option values and convert to string
                    processed_value = str(replace_bracket_placeholders(value.strip(), var_dict))
                    options[key.strip()] = processed_value

        # Create unique identifier
        identifier = f"__MEDIA_{replacement_count}__"
        replacement_count += 1

        # Store media object information using the MediaObject class
        media_objects[identifier] = MediaObject(type=media_type, path=processed_path, options=options)

        return identifier

    # Use verbose flag for multiline regex pattern
    regex = re.compile(pattern, re.VERBOSE)

    # First replace media placeholders
    processed_text = regex.sub(replace_match, text)

    # Then evaluate any remaining Python expressions in the text
    processed_text = replace_bracket_placeholders(processed_text, var_dict)

    return processed_text, media_objects



import sys

sys.path.append("src")

import pytest
from owlsight import (
    TextGenerationProcessorTransformers,
    TextGenerationProcessorOnnx,
    TextGenerationProcessorGGUF,
    MultiModalProcessorTransformers,
)
from owlsight.processors.helper_functions import select_processor_type


@pytest.mark.parametrize(
    "model_path, expected_result",
    [
        (r"nvidia/NVLM-D-72B", TextGenerationProcessorTransformers),  # Replace with the expected result
        (
            r"C:\Users\Test\.cache\huggingface\hub\models--microsoft--Phi-3-mini-4k-instruct-gguf",
            TextGenerationProcessorGGUF,
        ),
        (r"EleutherAI/gpt-neo-2.7B", TextGenerationProcessorTransformers),
        (
            r"C:\Users\Test\.cache\lm-studio\models\hugging-quants\Llama-3.2-1B-Instruct-Q8_0-GGUF",
            TextGenerationProcessorGGUF,
        ),
        (r"C:\Users\Test\.cache\huggingface\hub\models--yhavinga--t5-base-dutch", TextGenerationProcessorTransformers),
        (r"microsoft/Phi-3.5-mini-instruct-onnx", TextGenerationProcessorOnnx),
    ],
)
def test_select_processor_type(model_path, expected_result):
    assert select_processor_type(model_path) == expected_result


def test_select_media_processor_type(media_model_mappings):
    for task, model_id in media_model_mappings.items():
        result = select_processor_type(model_id, task)
        assert result == MultiModalProcessorTransformers


if __name__ == "__main__":
    pytest.main([__file__])



import sys
import pandas as pd
import numpy as np
from datetime import datetime, timedelta

sys.path.append("src")

import pytest

from owlsight.utils.helper_functions import replace_bracket_placeholders, parse_media_placeholders
from owlsight.utils.custom_classes import MediaObject

# Fixture for test data
@pytest.fixture
def test_context():
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": ["x", "y", "z"]})

    test_array = np.array([1, 2, 3])
    test_dict = {"name": "Alice", "scores": [85, 90, 95]}
    test_date = datetime(2024, 1, 1)

    return {
        "pd": pd,
        "np": np,
        "df": df,
        "test_array": test_array,
        "test_dict": test_dict,
        "test_date": test_date,
        "datetime": datetime,
        "timedelta": timedelta,
    }


# Basic expressions
@pytest.mark.parametrize(
    "test_id, input_string, expected",
    [
        ("basic_1", "{{1 + 1}}", 2),
        ("basic_2", "Result: {{2 * 3}}", "Result: 6"),
    ],
)
def test_basic_expressions(test_context, test_id, input_string, expected):
    result = replace_bracket_placeholders(input_string, test_context)
    assert result == expected


# Data structures
@pytest.mark.parametrize(
    "test_id, input_string, expected",
    [
        ("struct_1", "{{{1, 2, 3}}}", {1, 2, 3}),
        ("struct_2", "{{[x for x in range(3)]}}", [0, 1, 2]),
        ("struct_3", "{{{'key': 5}}}", {"key": 5}),
        ("struct_4", "{{(1, 2, 3)}}", (1, 2, 3)),
    ],
)
def test_data_structures(test_context, test_id, input_string, expected):
    result = replace_bracket_placeholders(input_string, test_context)
    assert result == expected


# String operations
@pytest.mark.parametrize(
    "test_id, input_string, expected",
    [
        ("str_1", "{{'hello'.upper()}}", "HELLO"),
        ("str_2", '{{", ".join(["a", "b", "c"])}}', "a, b, c"),
    ],
)
def test_string_operations(test_context, test_id, input_string, expected):
    result = replace_bracket_placeholders(input_string, test_context)
    assert result == expected


# complex operations
@pytest.mark.parametrize(
    "test_id, input_string, expected",
    [
        ("complex_1", "{{5*2}} is more than {{2*3}}", "10 is more than 6"),
        ("complex_2", '{{" ".join([str(x) for x in range(3)])}}', "0 1 2"),
    ],
)
def test_dict_operations(test_context, test_id, input_string, expected):
    result = replace_bracket_placeholders(input_string, test_context)
    assert result == expected


# Pandas operations
@pytest.mark.parametrize(
    "test_id, input_string, expected",
    [
        ("pd_1", '{{df["A"].mean()}}', 2.0),
        ("pd_2", '{{df["B"].max()}}', 6),
        ("pd_3", '{{df["C"].value_counts().to_dict()}}', {"x": 1, "y": 1, "z": 1}),
        ("pd_4", 'Average of A: {{df["A"].mean()}}', "Average of A: 2.0"),
        ("pd_5", '{{df.groupby("C")["A"].mean().to_dict()}}', {"x": 1.0, "y": 2.0, "z": 3.0}),
    ],
)
def test_pandas_operations(test_context, test_id, input_string, expected):
    result = replace_bracket_placeholders(input_string, test_context)
    assert result == expected


# Numpy operations
@pytest.mark.parametrize(
    "test_id, input_string, expected",
    [
        ("np_1", "{{np.mean(test_array)}}", 2.0),
        ("np_2", "{{np.sum(test_array)}}", 6),
    ],
)
def test_numpy_operations(test_context, test_id, input_string, expected):
    result = replace_bracket_placeholders(input_string, test_context)
    assert result == expected


# Date operations
@pytest.mark.parametrize(
    "test_id, input_string, expected",
    [
        ("date_1", "{{test_date.year}}", 2024),
        ("date_2", "{{test_date + timedelta(days=1)}}", datetime(2024, 1, 2)),
    ],
)
def test_date_operations(test_context, test_id, input_string, expected):
    result = replace_bracket_placeholders(input_string, test_context)
    assert result == expected


# Error handling
@pytest.mark.parametrize(
    "test_id, input_string, expected_error",
    [
        ("error_1", "{{undefined_variable}}", NameError),
        ("error_2", "{{1/0}}", ZeroDivisionError),
        ("error_3", '{{int("not a number")}}', ValueError),
    ],
)
def test_error_handling(test_id, input_string, expected_error):
    with pytest.raises(expected_error):
        replace_bracket_placeholders(input_string, {})


@pytest.mark.parametrize(
    "test_id, input_string, var_dict, expected_text, expected_media",
    [
        (
            "basic_image",
            "[[image:photo.jpg]]",
            {},
            "__MEDIA_0__",
            {"__MEDIA_0__": MediaObject(type="image", path="photo.jpg", options={})},
        ),
        (
            "basic_audio",
            "[[audio:recording.mp3]]",
            {},
            "__MEDIA_0__",
            {"__MEDIA_0__": MediaObject(type="audio", path="recording.mp3", options={})},
        ),
        (
            "basic_video",
            "[[video:clip.mp4]]",
            {},
            "__MEDIA_0__",
            {"__MEDIA_0__": MediaObject(type="video", path="clip.mp4", options={})},
        ),
    ],
)
def test_basic_media_syntax(test_id, input_string, var_dict, expected_text, expected_media):
    result_text, result_media = parse_media_placeholders(input_string, var_dict)
    assert result_text == expected_text
    assert result_media == expected_media

# Python expression integration
@pytest.mark.parametrize(
    "test_id, input_string, var_dict, expected_text, expected_media",
    [
        (
            "python_expr",
            "[[image:{{folder}}/{{filename}}]]",
            {"folder": "photos", "filename": "cat.jpg"},
            "__MEDIA_0__",
            {"__MEDIA_0__": MediaObject(type="image", path="photos/cat.jpg", options={})},
        ),
        (
            "python_expr_options",
            "[[image:{{folder}}/test.jpg|width={{size}}]]",
            {"folder": "images", "size": 512},
            "__MEDIA_0__",
            {"__MEDIA_0__": MediaObject(type="image", path="images/test.jpg", options={"width": "512"})},
        ),
    ],
)
def test_python_expression_integration(test_id, input_string, var_dict, expected_text, expected_media):
    result_text, result_media = parse_media_placeholders(input_string, var_dict)
    assert result_text == expected_text
    assert result_media == expected_media

# Multiple media objects and options
@pytest.mark.parametrize(
    "test_id, input_string, var_dict, expected_text, expected_media",
    [
        (
            "multiple_media",
            "Compare [[image:first.jpg]] with [[image:second.jpg]]",
            {},
            "Compare __MEDIA_0__ with __MEDIA_1__",
            {
                "__MEDIA_0__": MediaObject(type="image", path="first.jpg", options={}),
                "__MEDIA_1__": MediaObject(type="image", path="second.jpg", options={}),
            },
        ),
        (
            "complex_options",
            "[[image:photo.jpg|width=512|height=512|pipeline=depth-estimation]]",
            {},
            "__MEDIA_0__",
            {
                "__MEDIA_0__": MediaObject(
                    type="image",
                    path="photo.jpg",
                    options={"width": "512", "height": "512", "pipeline": "depth-estimation"},
                )
            },
        ),
    ],
)
def test_multiple_media_and_options(test_id, input_string, var_dict, expected_text, expected_media):
    result_text, result_media = parse_media_placeholders(input_string, var_dict)
    assert result_text == expected_text
    assert result_media == expected_media

# Mixed content
@pytest.mark.parametrize(
    "test_id, input_string, var_dict, expected_text, expected_media",
    [
        (
            "mixed_content",
            "The value is {{2 + 2}} and here's an [[image:test.jpg]]",
            {},
            "The value is 4 and here's an __MEDIA_0__",
            {"__MEDIA_0__": MediaObject(type="image", path="test.jpg", options={})},
        ),
        (
            "mixed_complex",
            """Process this [[image:{{folder}}/{{filename}}|width={{size}}]] 
            with value {{x + 1}} and [[audio:recording.mp3|language=en]]""",
            {"folder": "imgs", "filename": "test.jpg", "size": 256, "x": 5},
            """Process this __MEDIA_0__ 
            with value 6 and __MEDIA_1__""",
            {
                "__MEDIA_0__": MediaObject(type="image", path="imgs/test.jpg", options={"width": "256"}),
                "__MEDIA_1__": MediaObject(type="audio", path="recording.mp3", options={"language": "en"}),
            },
        ),
    ],
)
def test_mixed_content(test_id, input_string, var_dict, expected_text, expected_media):
    result_text, result_media = parse_media_placeholders(input_string, var_dict)
    assert result_text == expected_text
    assert result_media == expected_media

# Error handling
def test_invalid_media_type():
    with pytest.raises(ValueError):
        parse_media_placeholders("[[invalid:test.jpg]]", {})

def test_missing_path():
    with pytest.raises(ValueError):
        parse_media_placeholders("[[image:]]", {})

def test_invalid_option_format():
    with pytest.raises(ValueError):
        parse_media_placeholders("[[image:test.jpg|invalid_option]]", {})

if __name__ == "__main__":
    pytest.main([__file__, "-v"])


from typing import Any, Dict, List, Tuple
import json
import os
import traceback
from copy import deepcopy

from owlsight.utils.logger import logger
from owlsight.utils.constants import DEFAULTS, CHOICES
from owlsight.utils.helper_functions import flatten_dict
from owlsight.utils.validations import validate_key_is_nested_one_layer


class ConfigManager:
    """
    A singleton class which carries the configuration for the whole application.

    Most important to know, is that there are 2 different configurations:
    - self._config: the true configuration that is used in the application backend.
    - config_choices: the configuration that presented in the UI, where the user can toggle between choices.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ConfigManager, cls).__new__(cls)
            cls._instance.config = {}
        return cls._instance

    def __init__(self):
        """
        Initialize the configuration manager with default values.
        """
        self._defaults = deepcopy(DEFAULTS)
        self._choices = deepcopy(CHOICES)
        self._config = DottedDict(self._defaults)

    def get(self, key: str, default=None) -> Any:
        """
        Get a configuration value using dotted notation for nested keys.
        """
        keys = key.split(".")
        value = self._config
        for k in keys:
            if isinstance(value, dict):
                value = value.get(k, default)
            else:
                return default
        return value

    def set(self, key: str, value: Any) -> None:
        """
        Set a configuration value using dotted notation for nested keys.
        """
        keys = key.split(".")
        d = self._config
        for k in keys[:-1]:
            if k not in d:
                d[k] = {}  # Create the nested dictionary if it doesn't exist
            d = d[k]  # Move deeper into the nested dictionary
        d[keys[-1]] = value  # Set the final key's value

    def _get_toggle_choice(self, section: str, key: str) -> Any:
        """Helper method to prepare toggle choices for a given section and key."""
        return _prepare_toggle_choices(self._config[section][key], self._choices[section][key])

    def _get_basic_choice(self, section: str, key: str) -> Any:
        """Helper method to return a basic configuration choice."""
        return self._config[section][key]

    def _create_main_choices(self) -> Dict[str, Any]:
        """Create the config choices for the 'main' section."""
        return {
            "back": None,
            "max_retries_on_error": self._get_toggle_choice("main", "max_retries_on_error"),
            "prompt_code_execution": self._get_toggle_choice("main", "prompt_code_execution"),
            "prompt_retry_on_error": self._get_toggle_choice("main", "prompt_retry_on_error"),
            "extra_index_url": self._get_basic_choice("main", "extra_index_url"),
        }

    def _create_model_choices(self) -> Dict[str, Any]:
        """Create the config choices for the 'model' section."""
        return {
            "back": None,
            "model_id": self._get_basic_choice("model", "model_id"),
            "save_history": self._get_toggle_choice("model", "save_history"),
            "system_prompt": self._get_basic_choice("model", "system_prompt"),
            "transformers__device": self._get_toggle_choice("model", "transformers__device"),
            "transformers__quantization_bits": self._get_toggle_choice("model", "transformers__quantization_bits"),
            "transformers__stream": self._get_toggle_choice("model", "transformers__stream"),
            "transformers__use_fp16": self._get_toggle_choice("model", "transformers__use_fp16"),
            "gguf__filename": self._get_basic_choice("model", "gguf__filename"),
            "gguf__verbose": self._get_toggle_choice("model", "gguf__verbose"),
            "gguf__n_ctx": self._get_toggle_choice("model", "gguf__n_ctx"),
            "gguf__n_gpu_layers": self._get_toggle_choice("model", "gguf__n_gpu_layers"),
            "gguf__n_batch": self._get_toggle_choice("model", "gguf__n_batch"),
            "gguf__n_cpu_threads": self._get_toggle_choice("model", "gguf__n_cpu_threads"),
            "onnx__tokenizer": self._get_basic_choice("model", "onnx__tokenizer"),
            "onnx__verbose": self._get_toggle_choice("model", "onnx__verbose"),
            "onnx__num_threads": self._get_basic_choice("model", "onnx__num_threads"),
        }

    def _create_generate_choices(self) -> Dict[str, Any]:
        """Create the config choices for the 'generate' section."""
        return {
            "back": None,
            "stopwords": str(self._get_basic_choice("generate", "stopwords")),
            "max_new_tokens": self._get_toggle_choice("generate", "max_new_tokens"),
            "temperature": self._get_toggle_choice("generate", "temperature"),
            "generation_kwargs": str(self._get_basic_choice("generate", "generation_kwargs")),
        }

    def _create_rag_choices(self) -> Dict[str, Any]:
        """Create the config choices for the 'rag' section."""
        return {
            "back": None,
            "active": self._get_toggle_choice("rag", "active"),
            "target_library": self._get_basic_choice("rag", "target_library"),
            "top_k": self._get_toggle_choice("rag", "top_k"),
            "search_query": self._get_basic_choice("rag", "search_query"),
        }

    def _create_huggingface_choices(self):
        return {
            "back": None,
            "search": self._get_basic_choice("huggingface", "search"),
            "top_k": self._get_toggle_choice("huggingface", "top_k"),
            "task": self._get_toggle_choice("huggingface", "task"),
            "select_model": self._get_toggle_choice("huggingface", "select_model"),
        }

    @property
    def config_choices(self) -> Dict[str, Dict[str, Any]]:
        """
        Get the configuration choices for the UI.
        """
        return {
            "main": self._create_main_choices(),
            "model": self._create_model_choices(),
            "generate": self._create_generate_choices(),
            "rag": self._create_rag_choices(),
            "huggingface": self._create_huggingface_choices(),
        }

    def save(self, path: str) -> bool:
        """
        Save the configuration to a file as JSON.

        Parameters
        ----------
        path : str
            The path to save the configuration file.

        Returns
        -------
        bool
            True if the configuration was saved successfully, False otherwise.
        """
        err_msg = "Cannot save config."
        if not isinstance(path, str) or not path:
            logger.error(f"{err_msg} Invalid file path provided.")
            return False

        # Ensure that the directory exists
        directory = os.path.dirname(path)
        if directory and not os.path.exists(directory):
            logger.error(f"{err_msg} Directory does not exist: '{directory}'")
            return False

        try:
            with open(path, "w") as f:
                json.dump(
                    self._config,
                    f,
                    indent=4,
                )
                logger.info(f"Configuration saved successfully to '{path}'")
                return True
        except (IOError, OSError) as e:
            logger.error(f"{err_msg} Error writing to file '{path}': {e}")
            return False
        except TypeError as e:
            logger.error(f"{err_msg} Error serializing configuration to JSON: {e}")
            return False

    def load(self, path: str) -> bool:
        """
        Load the configuration from a file as JSON.
        It gets stored in the _config attribute.

        Parameters
        ----------
        path : str
            The path to the configuration file.

        Returns
        -------
        bool
            True if the configuration was loaded successfully, False otherwise.
        """
        err_msg = "Cannot load config."
        if not isinstance(path, str) or not path:
            logger.error("Invalid file path provided.")
            return False

        if not os.path.exists(path):
            logger.error(f"{err_msg} Configuration file does not exist: '{path}'")
            return False

        if not path.endswith(".json"):
            logger.error(f"{err_msg} Configuration file must be a JSON file.")
            return False

        try:
            with open(path, "r") as f:
                data = json.load(f)
        except (IOError, OSError) as e:
            logger.error(f"{err_msg} Error reading from file '{path}': {e}")
            return False
        except json.JSONDecodeError as e:
            logger.error(f"{err_msg} Error parsing JSON in file '{path}': {e}")
            return False

        try:
            config = DottedDict(data)
            self.validate_config(config)
            self._config = config
            logger.info(f"Configuration loaded successfully from '{path}'")
            return True
        except Exception:
            logger.error(f"{err_msg} Error loading configuration from '{path}': {traceback.format_exc()}")
            return False

    def validate_config(self, config: dict):
        """
        Validate the configuration.
        """
        flattened_defaults = flatten_dict(self._defaults)
        flattened_config = flatten_dict(config)

        # check differences in sections:
        missing_sections = set(self._defaults.keys()) - set(config.keys())
        if missing_sections:
            raise KeyError(f"Config misses the following sections: {missing_sections}")

        invalid_sections = set(config.keys()) - set(self._defaults.keys())
        if invalid_sections:
            raise KeyError(f"Config has the following sections, which are not valid in owlsight: {invalid_sections}")

        # check differences in keys:
        missing_keys = set(flattened_defaults.keys()) - set(flattened_config.keys())
        if missing_keys:
            raise KeyError(f"Config misses the following keys: {missing_keys}")

        invalid_keys = set(flattened_config.keys()) - set(flattened_defaults.keys())
        invalid_keys = {
            key for key in invalid_keys if validate_key_is_nested_one_layer(key)
        }  # only consider keys with nested keys
        if invalid_keys:
            raise KeyError(f"Config has the following keys, which are not valid in owlsight: {invalid_keys}")

        # check if values are valid
        flattened_choices = flatten_dict(self._choices)
        for key, value in flattened_config.items():
            if not validate_key_is_nested_one_layer(key):
                logger.warning(f"Key '{key}' is not nested one layer deep. Skipping validation.")
                continue
            choices = flattened_choices[key]
            if isinstance(choices, list) and choices != []:
                # if the value is a list, it means that the value is a togglechoice. we check if the value is in the list
                if value not in choices:
                    raise ValueError(f"Invalid value {value} for key '{key}'. Possible values: {choices}")
            else:
                # if the value is not a list, it means that the value is a basic choice. we check if the type is correct
                if not isinstance(value, type(choices)):
                    raise TypeError(
                        f"Invalid type {type(value).__name__} for key '{key}'. Expected type: {type(choices).__name__}"
                    )

    def __repr__(self):
        return repr(self._config)


class DottedDict(dict):
    """A dictionary with dotted access to attributes, enforcing lowercase keys."""

    def __getattr__(self, attr):
        attr = attr.lower()
        value = self.get(attr)
        if isinstance(value, dict):
            return DottedDict(value)  # Recursively return DottedDict for nested dicts
        return value

    def __setattr__(self, attr, value):
        self[attr.lower()] = value

    def __delattr__(self, attr):
        del self[attr.lower()]


def _dynamicly_initialize_toggle_choices(current_val: List[str], possible_vals: List[Any]) -> Tuple[str, List[Any]]:
    """
    In some cases, the possible values are not provided yet.
    This is if the possible values depend on some action which needs to be undertaken first, before the possible values can be determined.
    In this case, we dynamically initialize the possible values based on the current value.
    """
    first_item = current_val[0] if current_val else ""
    possible_vals = current_val
    current_val: str = first_item

    return current_val, possible_vals


def _prepare_toggle_choices(current_val: Any, possible_vals: List[Any]) -> List[Any]:
    """
    Prepare the config_choices to be used in the UI for toggling between choices.
    Parameters
    ----------
    current_val : Any
        The current value. Can be seen as default value.
    possible_vals : List[Any]
        The possible values for the configuration parameter.
        Allow user to toggle between the values.
    """
    # If the possible values are not provided, we assume that the current value is a list of possible values
    if not possible_vals and isinstance(current_val, list):
        current_val, possible_vals = _dynamicly_initialize_toggle_choices(current_val, possible_vals)

    if current_val in possible_vals:
        index = possible_vals.index(current_val)
        possible_vals = possible_vals[index:] + possible_vals[:index]

    return possible_vals
